from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

from os.path import join
from re import search
from sdss_access import AccessError
from sdss_access.sync.baseaccess import BaseAccess


class RsyncAccess(BaseAccess):
    """Class for providing Rsync access to SDSS SAS Paths
    """
    remote_scheme = 'rsync'
    access_mode = 'rsync'

    def __init__(self, label='sdss_rsync', stream_count=5, mirror=False, public=False, release=None, 
                 verbose=False):
        super(RsyncAccess, self).__init__(stream_count=stream_count, mirror=mirror, public=public, 
                                          release=release, verbose=verbose)
        self.label = label
        self.auth = None
        self.stream = None
        self.stream_count = stream_count
        self.verbose = verbose
        self.initial_stream = self.get_stream()

    def __repr__(self):
        return '<RsyncAccess(using="{0}")>'.format(self.netloc)

    def get_task_out(self, task=None):
        if task:
            command = "rsync -R %(source)s" % task
            if self.verbose:
                print(command)
            status, out, err = self.stream.cli.foreground_run(command)
            if status:
                raise AccessError("Return code %r\n%s" % (status, err))
        else:
            out = None
        return out

    def generate_stream_task(self, task=None, out=None):
        ''' creates the task to put in the download stream '''
        if task and out:
            depth = task['location'].count('/')
            if self.public:
                depth -= 1
            for result in out.split(b"\n"):
                result = result.decode('utf-8')
                if result.startswith(('d', '-', 'l')):
                    try:
                        location = search(r"^.*\s{1,3}(.+)$", result).group(1)
                    except Exception:
                        location = None
                    if location and location.count('/') == depth:
                        source = join(self.stream.source, location) if self.remote_base else None
                        destination = join(self.stream.destination, location)
                        yield (location, source, destination)

    def set_stream_task(self, task=None):
        out = self.get_task_out(task=task)
        super(RsyncAccess, self).set_stream_task(task=task, out=out)

    def _get_stream_command(self):
        ''' gets the stream command used when committing the download '''
        return "rsync -avRK --files-from={path} {source} {destination}"



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Filename: test_vacs.py
# Project: tools
# Author: Brian Cherinka
# Created: Sunday, 15th September 2019 7:56:52 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 15th September 2019 8:15:11 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pytest
from marvin import config
from marvin.tools.vacs import VACs


@pytest.fixture(scope='session')
def vacs():
    config.setRelease('DR15')
    v = VACs()
    yield v
    v = None


class TestVacs(object):

    def test_listvacs(self, vacs):
        assert len(vacs.list_vacs()) > 0

    def test_check_target(self, vacs):
        tdict = vacs.check_target('1-209232')
        assert tdict == {'firefly': True,
                         'galaxyzoo': True, 'gema': True, 'HI': False}

    def test_check_target_fail(self, vacs):
        with pytest.raises(AssertionError) as cm:
            tdict = vacs.check_target('8485-1901')
        assert cm.type == KeyError or cm.type == AssertionError
        assert 'Identifier "plateifu" is not available.  Try a "mangaid".' in str(
            cm.value)


class TestVACDataClass(object):
        
    def test_gzvac(self, vacs):
        gz = vacs.galaxyzoo
        assert gz.name == 'galaxyzoo'
        assert gz._path is not None
        assert gz.data is not None

    def test_has_target(self, vacs):
        gz = vacs.galaxyzoo
        assert gz.has_target('1-209232') == True

    def test_addmethod(self, vacs):
        hi = vacs.HI
        assert hasattr(hi, 'plot_mass_fraction')
        




from .base import *

from .MPL4 import MPL4
from .MPL5 import MPL5
from .MPL6 import MPL6
from .MPL7 import MPL7
from .MPL8 import MPL8
from .MPL9 import MPL9
from .MPL10 import MPL10

# Defines the list of datamodels.
datamodel = DAPDataModelList([MPL4, MPL5, MPL6, MPL7, MPL8, MPL9, MPL10])




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: utils.py
# Project: tree
# Author: Brian Cherinka
# Created: Tuesday, 17th March 2020 11:22:02 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Tuesday, 17th March 2020 1:41:46 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import os
import six
from tree import Tree


def compute_changelog(new, old, pprint=None, to_list=None, remove_sas=True, include_paths=True,
                      paths_only=None):
    ''' Compute the difference between two Tree environments

    Finds and prints the difference between two tree environment
    configurations.  Accepts either string names of config files, e.g.
    "dr16" and "dr15", or the preloaded `Tree` configs, e.g.
    `Tree(config='dr16')`.  By default returns a dictionary of changes or use
    the `pprint` keyword to return a formatted print-friendly string for display.
    The `to_list` keyword returns a print-formatted list of strings parse parseable
    by `docutree` for Sphinx docs.  Set `include_paths` to True to include a changelog
    of the `sdss_access` path definitions.

    Parameters:
        new (str|Tree):
            The new tree enviroment to compare
        old (str|Tree):
            The old tree environment to compare
        pprint (bool):
            If True, returns a single joined string for printing. Default is False.
        to_list (bool):
            If True, returns a list of strings formatted for printing. Default is False.
        remove_sas (bool):
            If True, removes the SAS_BASE_DIR from environment values.  Default is True.
        include_paths (bool):
            If True, includes changes to the sdss_access PATHS section.  Default is True.
        paths_only (bool):
            If True, returns only changes in sdss_access PATHS section.  Default is False.

    Returns:
        A dictionary of relevant changes between the two releases

    Example:
        >>> # print the differences DR16 and DR15
        >>> diffs = compute_changelog('dr16', 'dr15', pprint=True)
        >>> print(diffs)
    '''

    # compute the environment changelog dictionary
    cl_dict = compute_environment_changes(new, old, remove_sas=remove_sas)

    # compute and add the paths changelog dictionary
    if include_paths or paths_only:
        path_dict = compute_path_changes(new, old, prepend_header=paths_only)
        cl_dict.update(path_dict)

    # return PATHS section only
    if paths_only:
        cl_dict = cl_dict.copy()
        cl_dict.pop('environment')
        if pprint or to_list:
            return print_paths(cl_dict, to_string=not to_list, prepend_header=paths_only)
        return cl_dict

    if pprint or to_list:
        return print_environment(cl_dict, to_string=not to_list)

    return cl_dict


def compute_environment_changes(new, old, remove_sas=True):
    ''' Compute the difference between two Tree environments

    Compares two tree environment configurations and returns a
    dictionary with keys `new`, `removed`, and `changes` indicating
    newly added environments, newly removed environments, and
    environments where environment variable definitions have changed.
    Accepts either string names of config files, e.g.
    "dr16" and "dr15", or the preloaded `Tree` configs, e.g.
    `Tree(config='dr16')`.

    Parameters:
        new (str|Tree):
            The new tree enviroment to compare
        old (str|Tree):
            The old tree environment to compare
        remove_sas (bool):
            If True, removes the SAS_BASE_DIR from environment values.  Default is True.

    Returns:
        A dictionary of relevant changes between the two releases
    '''
    dd = {'environment': {'new': {}, 'changes': {}, 'removed': {}}}

    if isinstance(new, six.string_types):
        new = Tree(config=new.lower())

    if isinstance(old, six.string_types):
        old = Tree(config=old.lower())

    new_envs = new.environ.keys()
    old_envs = old.environ.keys()
    new_name = new.environ['default']['name']
    old_name = old.environ['default']['name']
    sas = os.getenv("SAS_BASE_DIR")

    dd['releases'] = {'new': new_name, 'old': old_name}

    # find unique new enviroments
    unique_new_envs = set(new_envs) - set(old_envs)
    for env in unique_new_envs:
        dd['environment']['new'][env] = {}
        for k, v in new.environ[env].items():
            if remove_sas:
                v = v.split(sas)[-1]
            dd['environment']['new'][env][k.upper()] = v

    # find unique old enviroments
    unique_old_envs = set(old_envs) - set(new_envs)
    for env in unique_old_envs:
        dd['environment']['removed'][env] = {}
        for k, v in old.environ[env].items():
            if remove_sas:
                v = v.split(sas)[-1]
            dd['environment']['removed'][env][k.upper()] = v

    # find new or updated environment variables
    for env in new_envs:
        # skip the default
        if env == 'default' or env not in old.environ:
            continue
        uniq_vars = set(new.environ[env].keys()) - set(old.environ[env].keys())
        if not uniq_vars:
            continue

        vbase = [v.split(os.path.join(sas, new_name))[-1] for v in new.environ[env].values()]
        voldbase = [v.split(os.path.join(sas, old_name))[-1] for v in old.environ[env].values()]
        changes = set(vbase) - set(voldbase)
        if not changes:
            continue

        dd['environment']['changes'][env] = {}
        for k, v in new.environ[env].items():
            vold = old.environ[env].get(k, None)
            if vold:
                vbase = v.split(os.path.join(sas, new_name))[-1]
                voldbase = vold.split(os.path.join(sas, old_name))[-1]
                if vbase != voldbase:
                    dd['environment']['changes'][env][k.upper()] = {'from': voldbase, 'to': vbase}
            else:
                if remove_sas:
                    v = v.split(sas)[-1]
                dd['environment']['changes'][env][k.upper()] = v

    return dd


def compute_path_changes(new, old, prepend_header=None):
    ''' Compute the difference between two Tree PATH sections

    Compares two tree PATH ini sections from the given environment
    configurations and returns  adictionary with keys `new`, and `updated`,
    indicating newly added paths, and any paths that have been modified
    from the last release.  Accepts either string names of config
    files, e.g. "dr16" and "dr15", or the preloaded `Tree` configs, e.g.
    `Tree(config='dr16')`.

    Parameters:
        new (str|Tree):
            The new tree enviroment to compare
        old (str|Tree):
            The old tree environment to compare
        prepend_header (bool):
            If True, adds a header to the diff with release versions

    Returns:
        A dictionary of relevant changes between the two releases
    '''

    dd = {'paths': {'new': {}, 'updated': {}}}

    if isinstance(new, six.string_types):
        new = Tree(config=new.lower())

    if isinstance(old, six.string_types):
        old = Tree(config=old.lower())

    new_templates = new.paths
    old_templates = old.paths

    # diffs = []
    if prepend_header and 'releases' not in dd:
        new_name = new.environ['default']['name']
        old_name = old.environ['default']['name']
        dd['releases'] = {'new': new_name, 'old': old_name}

    unique_new = sorted(set(new_templates) - set(old_templates))

    if unique_new:
        for path in unique_new:
            dd['paths']['new'][path] = new_templates[path]

    # changes
    changed = []
    for name, template in new_templates.items():
        if name in old_templates and template != old_templates[name]:
            changed.append((name, template, old_templates[name]))

    if changed:
        for item in changed:
            name, newt, oldt = item
            dd['paths']['updated'][name] = {'from': oldt, 'to': newt}

    return dd


def print_environment(changes, to_string=None):
    ''' Print the environment changelog

    Formats the changelog into a print-friendly list of
    strings.

    Parameters:
        changes (dict):
            A dictionary of changes between two releases
        to_string (bool):
            If True, returns a single string

    Returns:
        A string list of print-formatted changes
    '''

    new_name = changes['releases']['new']
    old_name = changes['releases']['old']

    diffs = ['Changes: {0} from {1}'.format(new_name.upper(), old_name.upper())]
    diffs.append(' ')

    # print unique new enviroments
    for env, values in changes['environment']['new'].items():
        diffs.append('New Environment: {0}'.format(env.upper()))
        diffs.append('----------------------')
        for key, val in values.items():
            diffs.append('{0}: {1}'.format(key, val))
        diffs.append(' ')

    # print unique old enviroments
    for env, values in changes['environment']['removed'].items():
        diffs.append('Removed Environment: {0}'.format(env.upper()))
        diffs.append('----------------------')
        for key, val in values.items():
            diffs.append('{0}: {1}'.format(key, val))
        diffs.append(' ')

    # print new or updated environment variables
    for env, values in changes['environment']['changes'].items():
        diffs.append('Changes in {0}:'.format(env.upper()))
        diffs.append('----------------------')
        for key, val in values.items():
            if isinstance(val, dict):
                diffs.append('Updated {0}: {1} from {2}'.format(key, val['to'], val['from']))
            else:
                diffs.append('New {0}: {1}'.format(key, val))
        diffs.append(' ')

    if 'paths' in changes:
        diffs.append('Changes in PATHS:')
        diffs.append('-----------------')
        path_changes = print_paths(changes)
        diffs.extend(path_changes)

    if to_string:
        diffs = '\n'.join(diffs)
    return diffs


def print_paths(changes, to_string=None, prepend_header=None):
    ''' Print the path changelog

    Formats the changelog into a print-friendly list of
    strings.

    Parameters:
        changes (dict):
            A dictionary of changes between two releases
        to_string (bool):
            If True, returns a single string

    Returns:
        A string list of print-formatted changes
    '''

    diffs = []
    if prepend_header and 'releases' in changes:
        new_name = changes['releases']['new']
        old_name = changes['releases']['old']
        diffs = ['Changes: {0} from {1}'.format(new_name.upper(), old_name.upper())]
        diffs.append(' ')

    # print unique new paths
    if changes['paths']['new']:
        diffs.append('New Paths:')
        diffs.append('----------')
        for name, template in changes['paths']['new'].items():
            diffs.append('{0}: {1}'.format(name, template))

    # changes
    if changes['paths']['updated']:
        diffs.append(' ')
        diffs.append('Updated Paths:')
        diffs.append('--------------')
        for name, vals in changes['paths']['updated'].items():
            diffs.append('{0}: '.format(name))
            diffs.append('   from: {0}'.format(vals['from']))
            diffs.append('   to: {0}'.format(vals['to']))

    if to_string:
        diffs = '\n'.join(diffs)

    return diffs



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brett Andrews
# @Date:   2017-04-11


from __future__ import print_function, division, absolute_import

import importlib
import six.moves
import pytest

import sdss_access
from sdss_access.sync.auth import Auth, AuthMixin
from sdss_access.path import AccessError


class TestAuth(object):

    def test_nonetrc_fails(self, monkeyhome):
        ''' test raise error when no netrc present '''
        with pytest.raises(AccessError) as cm:
            Auth()
        assert 'No netrc file found. Please create one.' in str(cm.value)

    def test_nonetrc_public_pass(self, monkeyhome):
        ''' test public access does not fail when no netrc '''
        auth = Auth(public=True)
        assert auth.public is True
        assert auth.ready() is None
        assert auth.netrc is None

    @pytest.mark.parametrize('netloc', [('data.sdss.org'), ('dtn01.sdss.org')])
    def test_setnetloc(self, netloc):
        ''' test setting a url domain location '''
        auth = Auth(public=True, netloc=netloc)
        assert auth.netloc == netloc

    def test_auth_goodnetrc(self):
        ''' test for a netrc loaded '''
        auth = Auth()
        assert auth.netrc is not None
        assert 'data.sdss.org' in auth.netrc.hosts.keys()

    @pytest.mark.parametrize('public', [(True), (False)])
    def test_auth_defaultuser(self, public):
        ''' test auth load default user '''
        auth = Auth(netloc='data.sdss.org', public=public)
        auth.load()
        if public:
            assert auth.username is None
        else:
            assert auth.username == auth.netrc.authenticators('data.sdss.org')[0]

    def test_auth_setuserpass(self):
        ''' test set userpass '''
        auth = Auth(netloc='data.sdss.org')
        assert auth.username is None
        assert auth.password is None
        auth.set_username('testuser')
        auth.set_password('testpass')
        assert auth.username == 'testuser'
        assert auth.password == 'testpass'

    @pytest.mark.parametrize('prompt, exp', [('blah', 'blah'), (None, 'sdss')])
    def test_user_inquire(self, monkeypatch, prompt, exp):
        ''' test username input prompt '''
        # mock the six.moves.input
        def mockinput(value):
            return prompt

        monkeypatch.setattr(six.moves, 'input', mockinput)
        # reload the auth module
        importlib.reload(sdss_access.sync.auth)
        from sdss_access.sync.auth import Auth

        # run the test
        auth = Auth()
        auth.set_username(inquire=True)
        assert auth.username == exp


class TmpAccess(AuthMixin, object):
    def __init__(self, public=None, netloc=None, verbose=None):
        self.public = public
        self.netloc = netloc
        self.verbose = verbose


class TestAuthMixin(object):

    def test_set_auth(self):
        ''' test the set_auth method '''
        ta = TmpAccess(netloc='data.sdss.org')
        ta.set_auth()
        assert ta.netloc == 'data.sdss.org'
        assert ta.auth is not None
        assert ta.auth.netloc == 'data.sdss.org'
        assert ta.auth.username == 'sdss'

    def test_set_auth_public(self):
        ''' test the public option '''
        ta = TmpAccess(netloc='data.sdss.org', public=True)
        ta.set_auth()
        assert ta.auth is not None
        assert ta.auth.public is True
        assert ta.auth.username is None




#!/usr/bin/env python
# encoding: utf-8
#
# conftest.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import shutil
import pytest


def make_dirs(path):
    (path / 'dr14/manga/spectro/redux/v2_1_2').mkdir(parents=True)
    (path / 'dr15/manga/spectro/redux/v2_4_3').mkdir(parents=True)
    (path / 'dr15/manga/spectro/analysis/v2_4_3/2.1.2').mkdir(parents=True)
    (path / 'dr15/manga/HI/v1_0_1').mkdir(parents=True)


@pytest.fixture()
def faketree(monkeypatch, tmp_path):
    # set up temp tree path
    p = tmp_path / "treetmp"
    p.mkdir(parents=True)
    # create fake datamodel directories
    make_dirs(p)
    monkeypatch.setenv('SAS_BASE_DIR', str(p))
    # setup fake config stuff
    treedir = p / 'treedir'
    treedir.mkdir(parents=True)
    # copy data over
    olddatadir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../data'))
    shutil.copytree(olddatadir, str(treedir / 'data'))
    (treedir / 'etc').mkdir()
    monkeypatch.setenv('TREE_DIR', str(treedir))
    # setup modulesdir
    mdir = p / 'modules'
    mdir.mkdir(parents=True)
    monkeypatch.setenv('MODULES_DIR', str(mdir))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: utils.py
# Project: path
# Author: Brian Cherinka
# Created: Thursday, 23rd April 2020 1:24:14 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Thursday, 23rd April 2020 1:24:15 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from tree import Tree
from tree.changelog import compute_changelog as compute_path_changes


def compute_changelog(new, old, pprint=None, to_list=None):
    ''' Compute the difference between two Tree PATH sections

    Compares two tree PATH ini sections from the given environment
    configurations and returns  adictionary with keys `new`, and `updated`,
    indicating newly added paths, and any paths that have been modified
    from the last release.  Accepts either string names of config
    files, e.g. "dr16" and "dr15", or the preloaded `Tree` configs, e.g.
    `Tree(config='dr16')`.

    Parameters:
        new (str|Tree):
            The new tree enviroment to compare
        old (str|Tree):
            The old tree environment to compare
        pprint (bool):
            If True, returns a single joined string for printing. Default is False.
        to_list (bool):
            If True, returns a list of strings formatted for printing. Default is False.

    Returns:
        A dictionary of relevant changes between the two releases
    '''

    diffs = compute_path_changes(new, old, pprint=pprint, paths_only=True, to_list=to_list)

    return diffs


def get_path_templates(name, public=None):
    ''' Return the path templates for all releases

    Produces a dictionary of path templates for a given
    path name for all releases.  Set public keyword to toggle only
    public data releases (DRs).

    Parameters:
        name (str):
            The sdss_access path name
        public (bool):
            If True, selects only public data releases

    Returns:
        A dictionary of the path template per release
    '''
    releases = Tree.get_available_releases(public=public)
    versions = {}
    for release in reversed(releases):
        release = 'sdsswork' if 'WORK' in release else release
        tree = Tree(config=release)
        versions[release] = tree.paths.get(name, None)
    return versions


def get_available_releases(public=None):
    ''' Get the available releases

    Parameters:
        public (bool):
            If True, only return public data releases
    '''
    return Tree.get_available_releases(public=public)



#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, division, print_function
from sdsstools import get_config, get_logger, get_package_version


NAME = 'sdss-tree'


# init the logger
log = get_logger(NAME)


# Loads config
config = get_config('tree')


__version__ = get_package_version(path=__file__, package_name=NAME)


from tree.tree import Tree



#!usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-14 14:31:36
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-14 19:03:32

from __future__ import print_function, division, absolute_import
import sys
import os
import argparse
import glob
import shutil
import time
import re
import importlib.util


def _remove_link(link):
    ''' Remove a symlink if it exists

    Parameters:
        link (str):
            The symlink filepath
    '''
    if os.path.islink(link):
        os.remove(link)


def make_symlink(src, link):
    '''create a symlink

    Parameters:
        src (str):
            The fullpath source of the symlink
        link (str):
            The symlink file path
    '''
    _remove_link(link)
    os.symlink(src, link)


def create_index_table(environ, envdir):
    ''' create an html table

    Parameters:
        environ (dict):
            A tree environment dictionary
        envdir (str):
            The filepath for the env directory
    Returns:
        An html table definition string
    '''

    table_header = """<table id="list" cellpadding="0.1em" cellspacing="0">
<colgroup><col width="55%"/><col width="20%"/><col width="25%"/></colgroup>
<thead>
    <tr><th><a href="?C=N&O=A">File Name</a>&nbsp;<a href="?C=N&O=D">&nbsp;&darr;&nbsp;</a></th><th><a href="?C=S&O=A">File Size</a>&nbsp;<a href="?C=S&O=D">&nbsp;&darr;&nbsp;</a></th><th><a href="?C=M&O=A">Date</a>&nbsp;<a href="?C=M&O=D">&nbsp;&darr;&nbsp;</a></th></tr>
</thead><tbody>
    <tr><td><a href="../">Parent directory/</a></td><td>-</td><td>-</td></tr>"""
    table_footer = """</tbody></table>"""

    # create table
    table = table_header

    # loop over the environment
    for section, values in environ.items():
        if section == 'default':
            continue

        for tree_name, tree_path in values.items():
            skipmsg = 'Skipping {0} for {1}'.format(tree_name, section)
            if '_ROOT' in tree_name:
                continue

            # create the src and target links
            src = tree_path
            link = os.path.join(envdir, tree_name.upper())

            # get the local time of the symlink
            try:
                stattime = time.strftime('%d-%b-%Y %H:%M', time.localtime(os.stat(src).st_mtime))
            except OSError:
                print("{0} does not appear to exist, skipping...".format(src))
                _remove_link(link)
                continue

            # skip the sas_base_dir
            if section == 'general' and 'SAS_BASE_DIR' in tree_name:
                print(skipmsg)
                continue

            # only create symlinks
            if section == 'general' and tree_name in ['CAS_LOAD', 'STAGING_DATA']:
                # only create links here if the target exist
                if os.path.exists(src):
                    make_symlink(src, link)
                else:
                    print(skipmsg)
            else:
                print('Processing {0} for {1}'.format(tree_name, section))
                make_symlink(src, link)

            # create the table entry
            if os.path.exists(link):
                table += '    <tr><td><a href="{0}/">{0}/</a></td><td>-</td><td>{1}</td></tr>\n'.format(tree_name.upper(), stattime)

    table += table_footer
    return table


def create_index_page(environ, defaults, envdir):
    ''' create the env index html page

    Builds the index.html page containing a table of symlinks
    to datamodel directories

    Parameters:
        environ (dict):
            A tree environment dictionary
        defaults (dict):
            The defaults dictionary from environ['default']
        envdir (str):
            The filepath for the env directory
    Returns:
        A string defintion of an html page
    '''

    # header of index file
    header = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head><meta name="viewport" content="width=device-width"/><meta http-equiv="content-type" content="text/html; charset=utf-8"/><style type="text/css">body,html {{background:#fff;font-family:"Bitstream Vera Sans","Lucida Grande","Lucida Sans Unicode",Lucidux,Verdana,Lucida,sans-serif;}}tr:nth-child(even) {{background:#f4f4f4;}}th,td {{padding:0.1em 0.5em;}}th {{text-align:left;font-weight:bold;background:#eee;border-bottom:1px solid #aaa;}}#list {{border:1px solid #aaa;width:100%%;}}a {{color:#a33;}}a:hover {{color:#e33;}}</style>
<link rel="stylesheet" href="{url}/css/sas.css" type="text/css"/>
<title>Index of /sas/{name}/env/</title>
</head><body><h1>Index of /sas/{name}/env/</h1>
"""

    # footer of index file
    footer = """<h3><a href='{url}/sas/'>{location}</a></h3>
<p>This directory contains links to the contents of
environment variables defined by the tree product, version {name}.
To examine the <em>types</em> of files contained in each environment variable
directory, visit <a href="/datamodel/files/">the datamodel.</a></p>
</body></html>
"""
    # create index html file
    index = header.format(**defaults)
    index += create_index_table(environ, envdir)
    index += footer.format(**defaults)

    return index


def create_env(environ, mirror=None, verbose=None):
    ''' create the env symlink directory structure

    Creates the env folder filled with symlinks to datamodel directories
    for a given tree config file.

    Parameters:
        environ (dict):
            A tree environment dictionary
        mirror (bool):
            If True, use the SAM url location
        verbose (bool):
            If True, print more information
    '''

    defaults = environ['default'].copy()
    defaults['url'] = "https://data.mirror.sdss.org" if mirror else "https://data.sdss.org"
    defaults['location'] = "SDSS-IV Science Archive Mirror (SAM)" if mirror else "SDSS-IV Science Archive Server (SAS)"

    if not os.path.exists(environ['general']['SAS_ROOT']):
        if verbose:
            print("{0} doesn't exist, skipping env link creation.".format(environ['general']['SAS_ROOT']))
        return

    if verbose:
        print("Found {0}.".format(environ['general']['SAS_ROOT']))

    # sets and creates envdir
    envdir = os.path.join(environ['general']['SAS_ROOT'], 'env')
    if not os.path.exists(envdir):
        os.makedirs(envdir)
    if not os.access(envdir, os.W_OK):
        return

    # create index html
    index = create_index_page(environ, defaults, envdir)

    # write the index file
    indexfile = os.path.join(envdir, 'index.html')
    with open(indexfile, 'w') as f:
        f.write(index)


def check_sas_base_dir(root=None):
    ''' Check for the SAS_BASE_DIR environment variable

    Will set the SAS_BASE_DIR in your local environment
    or prompt you to define one if is undefined

    Parameters:
        root (str):
            Optional override of the SAS_BASE_DIR envvar

    '''
    sasbasedir = root or os.getenv("SAS_BASE_DIR")
    if not sasbasedir:
        sasbasedir = input('Enter a path for SAS_BASE_DIR: ')
    os.environ['SAS_BASE_DIR'] = sasbasedir


def write_header(term='bash', tree_dir=None, name=None):
    ''' Write proper file header in a given shell format

    Parameters:
        term (str):
            The type of shell header to write, can be "bash", "tsch", or "modules"
        tree_dir (str):
            The path to this repository
        name (str):
            The name of the configuration

    Returns:
        A string header to insert
    '''

    assert term in ['bash', 'tsch', 'modules'], 'term must be either bash, tsch, or module'

    product_dir = tree_dir.rstrip('/')
    base = 'export' if term == 'bash' else 'setenv'
    sep = '=' if term == 'bash' else ' '

    if term != 'modules':
        hdr = """# Set up tree/{0} for {1}
{2} TREE_DIR{4}{3}
{2} TREE_VER{4}{0}
{2} PATH{4}$TREE_DIR/bin:$PATH
{2} PYTHONPATH{4}$TREE_DIR/python:$PYTHONPATH
                """.format(name, term, base, product_dir, sep)
    else:
        hdr = """#%Module1.0
proc ModulesHelp {{ }} {{
    global product version
    puts stderr "This module adds $product/$version to various paths"
}}
set name tree
set product tree
set version {1}
conflict $product

module load sdsstools/0.1.7
prereq sdsstools/0.1.7
module load sdss_access
prereq sdss_access

module-whatis "Sets up $product/$version in your environment"

set PRODUCT_DIR {0}
setenv [string toupper $product]_DIR $PRODUCT_DIR
setenv [string toupper $product]_VER $version
prepend-path PATH $PRODUCT_DIR/bin
prepend-path PYTHONPATH $PRODUCT_DIR/python

                """.format(product_dir, name)

    return hdr.strip()


def write_version(name):
    ''' Make the default modules version string '''
    modules_version = "#%Module1.0\nset ModulesVersion {0}".format(name)
    return modules_version


def write_file(environ, term='bash', out_dir=None, tree_dir=None, default=None):
    ''' Write a tree environment file

    Loops over the tree environ and writes them out to a bash, tsch, or
    modules file

    Parameters:
        environ (dict):
            The tree dictionary environment
        term (str):
            The type of shell header to write, can be "bash", "tsch", or "modules"
        tree_dir (str):
            The path to this repository
        out_dir (str):
            The output path to write the files (default is etc/)
        default (str):
            The default config to write into the .version file

    '''

    # get the proper name, header and file extension
    name = environ['default']['name']
    header = write_header(term=term, name=name, tree_dir=tree_dir)
    exts = {'bash': '.sh', 'tsch': '.csh', 'modules': '.module'}
    ext = exts[term]

    # shell command
    if term == 'bash':
        cmd = 'export {0}={1}\n'
    else:
        cmd = 'setenv {0} {1}\n'

    # write the environment config files
    filename = os.path.join(out_dir, name + ext)
    with open(filename, 'w') as f:
        f.write(header + '\n')
        for key, values in environ.items():
            if key != 'default':
                # write separator
                f.write('#\n# {0}\n#\n'.format(key))
                # write tree names and paths
                for tree_name, tree_path in values.items():
                    if tree_path.startswith(os.getenv("SAS_BASE_DIR")):
                        f.write(cmd.format(tree_name.upper(), tree_path))

    # write default .version file for modules (or default for lua modules)
    default = default if default else name
    modules_version = write_version(default)
    if term == 'modules' and environ['default']['current'] == 'True':
        # write .version file for tcl
        version_name = os.path.join(out_dir, '.version')
        with open(version_name, 'w') as f:
            f.write(modules_version)


def get_python_path():
    ''' Finds and switches to the tree python package directory '''
    # get the TREE directory
    tree_dir = os.getenv('TREE_DIR', None)
    if not tree_dir:
        path = os.path.dirname(os.path.abspath(__file__))
        tree_dir = os.path.realpath(os.path.join(path, '..'))
    pypath = os.path.join(tree_dir, 'python')

    if pypath not in sys.path:
        sys.path.append(pypath)
    os.chdir(pypath)


def get_tree(config=None):
    ''' Get the tree for a given config

    Parameters:
        config (str):
            The name of the tree config to load

    Returns:
        a Python Tree instance
    '''

    # ensure the tree package is importable
    mod = importlib.util.find_spec('tree')
    if not (mod and mod.origin):
        get_python_path()

    # extract the config format from either XXXX.cfg or full filepath
    has_cfg = re.search(r'(\w+)\.cfg', config)
    if has_cfg:
        config = has_cfg.group()
    from tree.tree import Tree
    tree = Tree(config=config)
    return tree


def copy_modules(filespath=None, modules_path=None, verbose=None, default=None):
    ''' Copy over the tree module files into your path '''

    # find or define a modules path
    if not modules_path:
        modulepath = os.getenv("MODULEPATH")
        if not modulepath:
            modules_path = input('Enter the root path for your module files:')
        else:
            split_mods = modulepath.split(':')
            if len(split_mods) > 1:
                # select which module paths to use
                items = ['Multiple module paths found. Choose which module paths to use '
                         '(e.g. "1,2") or hit enter for "all". Or "q" to quit: ']
                items += ['{0}. {1}'.format(i + 1, t) for i, t in enumerate(split_mods)] + ['\n']
                msg = '\n'.join(items)
                selected = input(msg) or 'all'
                # check to quit
                if selected == 'q':
                    if verbose:
                        print('Quitting module copy.')
                    return

                # select choices
                choices = range(len(split_mods)) if selected == 'all' else [
                    int(i) - 1 for i in selected.split(',')]

                # loop over selected module paths to install tree
                for choice in choices:
                    mfile = split_mods[choice]
                    # run rest of method
                    copy_modules(filespath=filespath, modules_path=mfile, verbose=verbose,
                                 default=default)

                # exit the function after loop is over
                return
            else:
                modules_path = split_mods[0]
                if verbose:
                    print('Installing tree modules into {0}'.format(modules_path))

    # check for the tree module directory
    tree_mod = os.path.join(modules_path, 'tree')
    if not os.path.isdir(tree_mod):
        if verbose:
            print('Creating module tree directory: {0}'.format(tree_mod))
        os.makedirs(tree_mod)
    else:
        doit = input('{0} already exists! Overwrite? (y/n) \n'.format(tree_mod)) or 'n'
        if doit == 'n':
            return

    # copy the modules into the tree
    if verbose:
        print('Copying modules from {1} into {0}'.format(tree_mod, filespath))
    module_files = glob.glob(os.path.join(filespath, '*.module'))
    for mfile in module_files:
        base = os.path.splitext(os.path.basename(mfile))[0]
        tree_out = os.path.join(tree_mod, base)
        shutil.copy2(mfile, tree_out)

    # copy the default .version into the tree
    version = os.path.join(filespath, '.version')
    if os.path.isfile(version):
        shutil.copy2(version, tree_mod)

    # write default file for lua
    if default:
        defpath = os.path.join(tree_mod, 'default')
        realpath = default
        if os.path.islink(defpath):
            os.unlink(defpath)
        os.symlink(realpath, defpath)


def check_output_dir(output_dir):
    ''' Check the output directory '''

    # check if output directory is within a pip package directory
    if 'site-packages' in output_dir or output_dir.endswith('python/tree/etc'):
        output_dir = os.path.expanduser('~/.tree/environments')

    # create directory if it does not exist
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    return output_dir


def get_parser():
    ''' Parse the arguments '''

    description = ('Create bash, tsch, and module environment configuration files '
                   'defining relevant variables for accessing data products on '
                   'the SDSS Science Archive Server (SAS)')
    parser = argparse.ArgumentParser(prog='setup_tree.py', description=description)
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help='Print extra information.', default=False)
    parser.add_argument('-r', '--root', action='store', dest='root', default=os.getenv('SAS_BASE_DIR'),
                        help='Override the environment variable $SAS_BASE_DIR.', metavar='SAS_BASE_DIR')
    parser.add_argument('-t', '--treedir', action='store', dest='treedir', default=os.getenv('TREE_DIR'),
                        help='Override the environment variable $TREE_DIR.', metavar='TREE_DIR')
    parser.add_argument('-m', '--modulesdir', action='store', dest='modulesdir', default=os.getenv('MODULES_DIR'),
                        help='Your modules directory.  Defaults to $MODULES_DIR', metavar='MODULES_DIR')
    parser.add_argument('-e', '--env', action='store_true', dest='env',
                        help='Create tree environment symlinks.', default=False)
    parser.add_argument('-i', '--mirror', action='store_true', dest='mirror',
                        help='Use the mirror site (SAM) instead.')
    parser.add_argument('-o', '--only', action='store', dest='only', metavar='[xxx].cfg',
                        default=None, help='create links for only the specified tree config.')
    parser.add_argument('-d', '--default', action='store', dest='default', default='sdsswork',
                        help='Default config version to write into the .version file. Defaults to "sdsswork"')
    parser.add_argument('-p', '--path', action='store', dest='path', default=None,
                        help='Custom output path to copy environment files')

    return parser


def main(args):

    # parse arguments
    opts = get_parser().parse_args()

    # check for a treedir; if none found, set path to the parent directory
    if not opts.treedir:
        # look for installed python module
        mod = importlib.util.find_spec('tree')
        if mod and mod.origin:
            opts.treedir = os.path.dirname(mod.origin)
        else:
            opts.treedir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
        os.environ['TREE_DIR'] = opts.treedir

    # get directories
    datadir = os.path.join(opts.treedir, 'data')
    etcdir = opts.path if opts.path else os.path.join(opts.treedir, 'etc')
    etcdir = check_output_dir(etcdir)

    if opts.verbose:
        print('TREE_DIR: ', opts.treedir)
        print("Data Directory : ", datadir)
        print('Output Directory: ', etcdir)

    # config files
    configs = glob.glob(os.path.join(datadir, '*.cfg'))
    if not configs:
        print('No config files found in {0}.  Cannot proceed with tree setup. '
              'Check for correct TREE_DIR.'.format(datadir))
        return

    # check for the SAS_BASE_DIR
    check_sas_base_dir(root=opts.root)

    # Read and write the configuration files
    for cfgfile in configs:
        # skip basework config
        if 'basework' in cfgfile:
            continue

        tree = get_tree(config=cfgfile)
        # create env symlinks or write out tree module/bash files
        if opts.env:
            # skip creating the environ if a specific config if specified
            if opts.only and opts.only not in cfgfile:
                continue
            create_env(tree.environ, mirror=opts.mirror)
        else:
            write_file(tree.environ, term='modules', out_dir=etcdir, tree_dir=opts.treedir, default=opts.default)
            write_file(tree.environ, term='bash', out_dir=etcdir, tree_dir=opts.treedir)
            write_file(tree.environ, term='tsch', out_dir=etcdir, tree_dir=opts.treedir)

    # check for files
    module_files = glob.glob(os.path.join(etcdir, '*.module'))
    if not any(module_files):
        print('No module files created in {0}'.format(etcdir))
    else:
        print('Environment files created in: {0}'.format(etcdir))

    # Setup the modules
    print('Copying module files...')
    copy_modules(filespath=etcdir, modules_path=opts.modulesdir, verbose=opts.verbose, default=opts.default)


if __name__ == '__main__':
    main(sys.argv[1:])



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-12-05 10:33:59
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-12-05 10:53:56

from __future__ import print_function, division, absolute_import
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils import statemachine
import traceback
import importlib
import re
from sdss_access.path.changelog import get_available_releases

# This file is a basic example of how to write a Sphinx plugin to document
# custom data.


def _indent(text, level=1):
    ''' Does a proper indenting for Sphinx rst '''

    prefix = ' ' * (4 * level)

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if line.strip() else line)

    return ''.join(prefixed_lines())


def _format_templates(name, command, templates):
    ''' Creates a list-table directive

    for a set of defined environment variables

    Parameters:
        name (str):
            The name of the config section
        command (object):
            The sdss_access path instance
        templates (dict):
            A dictionary of the path templates

    Yields:
        A string rst-formated list-table directive

    '''

    yield '.. list-table:: {0} path definitions'.format(name)
    yield _indent(':widths: 20 100 70')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Template')
    yield _indent('  - Kwargs')
    for key, var in templates.items():
        kwargs = command.lookup_keys(key)
        yield _indent('* - {0}'.format(key))
        yield _indent('  - {0}'.format(var))
        yield _indent('  - {0}'.format(', '.join(kwargs)))
    yield ''


def _format_changelog(changes):
    ''' Format a changelog for a sdss_access Paths '''

    yield '.. role:: maroon'

    # yield lines from the PATHS section

    # list mode
    for line in changes:
        if 'Changes' in line:
            yield f'**{line}**'
            yield ''
        elif 'New' in line or 'Updated' in line:
            yield '* ' + _indent(f'**{line.strip()}**')
        elif re.search('[a-z]', line):
            # nested list From To
            if line.strip().startswith(('from', 'to')):
                line = line.replace('to:', '**To:**').replace('from:', '**From:**')
                yield _indent('    * ' + line, level=2)
            else:
                name, template = line.split(':', 1)
                yield _indent(f'    * :maroon:`{name}`: {template}')


def load_module(module_path, error=None, products=None):
    """Load the module."""

    # Exception to raise
    error = error if error else RuntimeError

    # __import__ will fail on unicode,
    # so we ensure module path is a string here.
    module_path = str(module_path)
    try:
        module_name, attr_name = module_path.split(':', 1)
    except ValueError:  # noqa
        raise error('"{0}" is not of format "module:object"'.format(module_path))

    # import the module
    try:
        mod = importlib.import_module(module_name)
    except (Exception, SystemExit) as exc:
        err_msg = 'Failed to import module "{0}". '.format(module_name)
        if isinstance(exc, SystemExit):
            err_msg += 'The module appeared to call sys.exit()'
        else:
            err_msg += 'The following exception was raised:\n{0}'.format(
                traceback.format_exc())

        raise error(err_msg)

    # check what kind of module
    if products:
        module = mod.dm.products
    else:
        module = mod

    if not hasattr(module, attr_name):
        raise error('Module "{0}" has no attribute "{1}"'.format(module_name, attr_name))

    return getattr(module, attr_name)


class PathDirective(rst.Directive):
    ''' The directive which instructs Sphinx how to format the Tree config documentation '''

    has_content = False
    required_arguments = 1
    option_spec = {
        'prog': directives.unchanged_required,
        'templates': directives.flag,
    }

    def _generate_nodes(self, name, command, templates=None):
        """Generate the relevant Sphinx nodes.

        Generates a section for the Tree datamodel.  Formats a tree section
        as a list-table directive.

        Parameters:
            name (str):
                The name of the config to be documented, e.g. 'sdsswork'
            command (object):
                The loaded module
            templates (bool):
                If True, generate a section for the path templates

        Returns:
            A section docutil node

        """

        # the source name
        source_name = name

        # Title
        section = nodes.section(
            '',
            nodes.title(text=name),
            ids=[nodes.make_id(name)],
            names=[nodes.fully_normalize_name(name)])

        # Summarize
        result = statemachine.ViewList()

        if templates:
            lines = _format_templates(name, command, command.templates)

        for line in lines:
            result.append(line, source_name)

        self.state.nested_parse(result, 0, section)

        return [section]

    def run(self):
        self.env = self.state.document.settings.env

        command = load_module(self.arguments[0])

        if 'prog' in self.options:
            prog_name = self.options.get('prog')
        else:
            raise self.error(':prog: must be specified')

        # get instance
        path = command(release=prog_name, public='DR' in prog_name)

        # get options
        templates = 'templates' in self.options

        # add a new section for paths
        return self._generate_nodes(prog_name, path, templates=templates)


class PathChangeDirective(rst.Directive):
    has_content = False
    required_arguments = 1
    option_spec = {
        'prog': directives.unchanged_required,
        'title': directives.unchanged,
        'drs': directives.unchanged_required,
    }

    def run(self):
        self.env = self.state.document.settings.env

        command = load_module(self.arguments[0])

        if 'prog' in self.options:
            prog_name = self.options.get('prog')
        else:
            raise self.error(':prog: must be specified')

        # get drs
        drs = self.options.get('drs').replace(' ', '')
        if drs == 'latest':
            releases = get_available_releases(public=True)
            old, new = releases[-2:]
        else:
            new, old = drs.split(',')

        # compute and format the changelog lines
        lines = command(new, old, to_list=True)
        title = lines[0]
        lines = lines[1:]
        lines = _format_changelog(lines)

        # the source name
        source_name = title

        # Title
        section = nodes.section(
            '',
            nodes.title(text=source_name),
            ids=[nodes.make_id(source_name)],
            names=[nodes.fully_normalize_name(source_name)])

        result = statemachine.ViewList()
        for line in lines:
            result.append(line, source_name)

        self.state.nested_parse(result, 0, section)

        return [section]


def setup(app):
    app.add_directive('datamodel', PathDirective)
    app.add_directive('changelog', PathChangeDirective)






from __future__ import absolute_import, division, print_function, unicode_literals

# The line above will help with 2to3 support.
from sdss_access.sync import CurlAccess, RsyncAccess
from sdss_access import is_posix

Base = RsyncAccess if is_posix else CurlAccess
access_mode = 'rsync' if is_posix else 'curl'
label = 'sdss_{0}'.format(access_mode)


class Access(Base):
    """Class for providing Rsync or Curl access depending on posix
    """
    access_mode = access_mode

    def __repr__(self):
        return '<Access(access_mode="{0}", using="{1}")>'.format(self.access_mode, self.netloc)




#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals
import os
from sdsstools import get_config, get_logger, get_package_version
from tree import Tree


# check if posix-based operating system
is_posix = (os.name == "posix")


NAME = 'sdss_access'

# init the logger
log = get_logger(NAME)

# set up the TREE, but match the TREE_VER if it is already there
config = os.environ.get('TREE_VER', 'sdsswork')
tree = Tree(config=config)
log.debug("SDSS_ACCESS> Using {0}".format(tree))


# Loads config
config = get_config(NAME)


#__version__ = '1.0.0dev'
__version__ = get_package_version(path=__file__, package_name=NAME)


from .path import Path as SDSSPath, AccessError
from .sync import HttpAccess, Access, BaseAccess, RsyncAccess, CurlAccess



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: test_access.py
# Project: sync
# Author: Brian Cherinka
# Created: Thursday, 8th August 2019 6:08:07 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Tuesday, 12th November 2019 2:56:34 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pytest
from sdss_access.sync import CurlAccess, RsyncAccess
import sdss_access
import sys
if sys.version_info.major == 2:
    import imp as importlib
else:
    import importlib


is_posix = [True, False]
exp = {True: (RsyncAccess, 'rsync'), False: (CurlAccess, 'curl')}


@pytest.fixture(params=is_posix, ids=["is_posix", "is_not_posix"])
def monkey_posix(monkeypatch, request):
    monkeypatch.setattr(sdss_access, 'is_posix', request.param)


def test_access(monkey_posix):
    Access = importlib.reload(sdss_access.sync.access).Access
    core, mode = exp[sdss_access.is_posix]
    assert Access.access_mode == mode
    assert issubclass(Access, core)




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: test_http.py
# Project: sync
# Author: Brian Cherinka
# Created: Monday, 16th September 2019 2:51:23 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Monday, 16th September 2019 3:08:02 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import os
import pytest
from sdss_access import tree, AccessError
from sdss_access.sync import HttpAccess


@pytest.fixture()
def monkeyprod(tmpdir, monkeypatch):
    ''' fixture to monkeypatch the sas_base_dir '''
    orig = os.getenv("PRODUCT_ROOT")
    tmppath = tmpdir / 'software'
    os.makedirs(tmppath, exist_ok=True)
    monkeypatch.setenv('PRODUCT_ROOT', str(tmppath))
    tree.set_product_root(root=str(tmppath))
    yield
    os.environ["PRODUCT_ROOT"] = orig
    tree.set_product_root()


class TestHttp(object):

    def test_http(self, http, datapath):
        name = http.release.lower() if http.release and 'dr' in http.release.lower() else datapath['work']
        path = http.url(datapath['name'], **datapath['params'])
        assert datapath['location'] in path
        assert 'https://data.sdss.org' in path
        assert name in path

    def test_svn_exist(self):
        http = HttpAccess(release='DR15')
        http.remote()
        exists = http.exists('mangapreimg', designid=8405, designgrp='D0084XX',
                             mangaid='1-42007', remote=True)
        assert exists is True

    def test_svn_get(self, monkeyprod):
        http = HttpAccess(release='DR15')
        http.remote()
        full = http.full('mangapreimg', designid=8405, designgrp='D0084XX',
                         mangaid='1-42007', remote=True)

        http.get('mangapreimg', designid=8405, designgrp='D0084XX',
                 mangaid='1-42007', remote=True)

        assert os.path.exists(full)

    def test_http_public_dl(self, monkeyprod):
        http = HttpAccess(release='DR14')
        http.remote()
        full = http.full('spec-lite', run2d='v5_10_0', plateid=3606, mjd=55182, fiberid=537)
        http.get('spec-lite', run2d='v5_10_0', plateid=3606, mjd=55182, fiberid=537)
        assert os.path.exists(full)

    def test_nonetrc_fails(self, monkeyhome):
        ''' test raise error when no netrc present '''
        with pytest.raises(AccessError) as cm:
            http = HttpAccess()
            http.remote()
        assert 'No netrc file found. Please create one.' in str(cm.value)

    def test_nonetrc_public_pass(self, monkeyhome):
        ''' test public access does not fail when no netrc '''
        http = HttpAccess(release='DR14')
        http.remote()
        assert http.public is True
        assert http.auth.username is None
        assert http.auth.ready() is None

    @pytest.mark.parametrize('tree_ver, exp', [('sdsswork', 'work'), ('dr15', 'dr15'),
                                               ('dr13', 'dr13'), ('mpl8', 'work')])
    def test_release_from_module(self, monkeypatch, tree_ver, exp, datapath):
        monkeypatch.setenv('TREE_VER', tree_ver)
        http = HttpAccess()
        full = http.full(datapath['name'], **datapath['params'])
        assert http.release == tree_ver
        assert exp in full




# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-11-29 10:28:58
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-27 17:22:02

from __future__ import print_function, division, absolute_import
import pytest
from tree.tree import Tree
import six
import os


@pytest.fixture()
def tree():
    t = Tree()
    t.replant_tree()
    yield t
    t = None


class TestTree(object):

    def test_load(self, tree):
        assert tree.treedir is not None
        assert isinstance(tree.treedir, six.string_types)
        assert tree.environ is not None
        assert 'APO' in tree.environ
        assert 'sdsswork' == tree.environ['default']['name']

    @pytest.mark.parametrize('keys', [('manga'), ('MANGA'), (['manga', 'apogee'])],
                             ids=['manga', 'capmanga', 'apomanga'])
    def test_load_keys(self, keys):
        t = Tree(key=keys)
        assert 'APO' not in t.environ
        assert 'default' in t.environ
        assert 'general' in t.environ
        keys = keys if isinstance(keys, list) else [keys]
        for key in keys:
            assert key.upper() in t.environ

    @pytest.mark.parametrize('dr', [('dr15')])
    def test_load_dr_config(self, dr):
        t = Tree(config=dr)
        assert dr == t.environ['default']['name']

    def _assert_paths(self, tree, dr=None):
        assert 'MANGA_SPECTRO_REDUX' in tree.environ['MANGA']
        assert 'MANGA_SPECTRO_ANALYSIS' in tree.environ['MANGA']
        if dr:
            assert dr in tree.environ['default']['name']
            assert dr in tree.environ['MANGA']['MANGA_ROOT']
            assert dr in tree.environ['MANGA']['MANGA_SPECTRO_REDUX']
            assert dr in tree.environ['MANGA']['MANGA_SPECTRO_ANALYSIS']

    @pytest.mark.parametrize('dr', [('dr15')])
    def test_load_with_update(self, tree, dr):
        assert 'sdsswork' in tree.environ['default']['name']
        assert 'mangawork' in tree.environ['MANGA']['MANGA_ROOT']
        tree = Tree(config=dr, update=True)
        self._assert_paths(tree, dr=dr)

    @pytest.mark.parametrize('dr', [('dr15')])
    def test_replant_tree(self, tree, dr):
        assert 'sdsswork' in tree.environ['default']['name']
        assert 'mangawork' in tree.environ['MANGA']['MANGA_ROOT']
        tree.replant_tree(dr)
        self._assert_paths(tree, dr=dr)

    def test_custom_paths_remain(self, monkeypatch):
        # set custom environ variable
        name = 'MANGA_QUICK'
        path = '/tmp/work/manga/quick'
        monkeypatch.setitem(os.environ, name, path)
        assert os.environ[name] == path
        tree = Tree()
        assert os.environ[name] == path
        assert path != tree.environ['MANGA'][name]

    def test_custom_paths_donotupdate(self, monkeypatch, tree):
        name = 'MANGA_QUICK'
        path = '/tmp/work/manga/quick'
        monkeypatch.setitem(os.environ, name, path)
        assert os.environ[name] == path
        tree.replant_tree('sdsswork')
        assert os.environ[name] != path
        assert os.environ[name] == tree.environ['MANGA'][name]

    def test_custom_paths_update_with_exclude(self, monkeypatch, tree):
        name = 'MANGA_QUICK'
        path = '/tmp/work/manga/quick'
        monkeypatch.setitem(os.environ, name, path)
        assert os.environ[name] == path
        tree.replant_tree('sdsswork', exclude=['MANGA_QUICK'])
        assert os.environ[name] == path
        assert path != tree.environ['MANGA'][name]

    def test_paths(self, tree):
        assert 'PATHS' not in tree.environ
        assert tree.paths is not None
        assert isinstance(tree.paths, dict)

    def test_get_orig_os(self, tree):
        # need to check if on system with no sdsswork modules running (i.e. Travis)
        orig_os = tree.get_orig_os_environ()
        hassas = orig_os.get("SAS_ROOT", None)
        if hassas:
            assert 'mangawork' in orig_os.get("MANGA_ROOT")
        else:
            assert 'MANGA_ROOT' not in orig_os

        assert 'mangawork' in os.getenv("MANGA_ROOT")
        tree.replant_tree('dr15')
        if hassas:
            assert 'dr15' not in orig_os.get("MANGA_ROOT")
        else:
            assert 'MANGA_ROOT' not in orig_os
        assert 'dr15' in os.getenv("MANGA_ROOT")

    def test_reset_orig_os(self, tree):
        # need to check if on system with no sdsswork modules running (i.e. Travis)
        orig_os = tree.get_orig_os_environ()
        hassas = orig_os.get("SAS_ROOT", None)
        tree.replant_tree('dr15')
        assert 'dr15' in os.getenv("MANGA_ROOT")
        tree.reset_os_environ()
        if hassas:
            assert 'mangawork' in os.getenv("MANGA_ROOT")
        else:
            assert "MANGA_ROOT" not in os.environ

    def test_list_configs(self, tree):
        cfgs = tree.list_available_configs()
        assert 'dr15.cfg' in cfgs

    @pytest.mark.parametrize('public', [(True), (False)])
    def test_get_releases(self, tree, public):
        rels = tree.get_available_releases(public=public)
        assert 'DR15' in rels
        if public:
            assert 'WORK' not in rels
        else:
            assert 'WORK' in rels

    @pytest.mark.parametrize('collapse', [(True), (False)])
    def test_to_dict(self, tree, collapse):
        envdict = tree.to_dict(collapse=collapse)
        if collapse:
            assert 'MANGA' not in envdict
            assert 'MANGA_ROOT' in envdict
        else:
            assert 'MANGA' in envdict
            assert 'MANGA_ROOT' not in envdict
            assert 'MANGA_ROOT' in envdict['MANGA']

    def test_set_product_root(self, tmp_path, tree):
        temp_prod = tmp_path / 'software'
        temp_prod.mkdir()
        tree.set_product_root(root=str(temp_prod))
        assert tree.productroot_dir == str(temp_prod)
        assert os.getenv("PRODUCT_ROOT") == str(temp_prod)

    def test_write_old_paths(self, faketree):

        tree = Tree()

        pathsfile = os.path.join(tree.treedir, 'data/sdss_paths.ini')
        assert not os.path.exists(pathsfile)

        tree.write_old_paths_inifile()

        assert os.path.exists(pathsfile)
        assert os.path.isfile(pathsfile)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-08-15
# @Filename: test_modelcube.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-08 15:08:26


from __future__ import absolute_import, division, print_function

import os

import pytest
import six
import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS

from marvin.core.exceptions import MarvinError
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.modelcube import ModelCube

from .. import marvin_test_if


@pytest.fixture(autouse=True)
def skipmpl4(galaxy):
    if galaxy.release == 'MPL-4':
        pytest.skip('No modelcubes in MPL-4')


def is_string(string):
    return isinstance(string, six.string_types)


class TestModelCubeInit(object):

    def _test_init(self, model_cube, galaxy, bintype=None, template=None):
        assert model_cube.release == galaxy.release
        assert model_cube._drpver == galaxy.drpver
        assert model_cube._dapver == galaxy.dapver
        assert model_cube.bintype.name == bintype if bintype else galaxy.bintype.name
        assert model_cube.template == template if template else galaxy.template
        assert model_cube.plateifu == galaxy.plateifu
        assert model_cube.mangaid == galaxy.mangaid
        assert isinstance(model_cube.header, fits.Header)
        assert isinstance(model_cube.wcs, WCS)
        assert model_cube._wavelength is not None
        assert model_cube._redcorr is not None

    def test_init_modelcube(self, galaxy, data_origin):
        if data_origin == 'file':
            kwargs = {'filename': galaxy.modelpath}
        elif data_origin == 'db':
            kwargs = {'plateifu': galaxy.plateifu}
        elif data_origin == 'api':
            kwargs = {'plateifu': galaxy.plateifu, 'mode': 'remote'}

        model_cube = ModelCube(**kwargs)
        assert model_cube.data_origin == data_origin
        assert model_cube.nsa.z == pytest.approx(galaxy.redshift)
        self._test_init(model_cube, galaxy)

    def test_init_from_file_global_mpl4(self, galaxy):
        model_cube = ModelCube(filename=galaxy.modelpath, release='MPL-4')
        assert model_cube.data_origin == 'file'
        self._test_init(model_cube, galaxy)

    def test_raises_exception_mpl4(self, galaxy):
        with pytest.raises(MarvinError) as cm:
            ModelCube(plateifu=galaxy.plateifu, release='MPL-4')
        assert 'ModelCube requires at least dapver=\'2.0.2\'' in str(cm.value)

    def test_init_modelcube_bintype(self, galaxy, data_origin):
        kwargs = {'bintype': galaxy.bintype.name}
        if data_origin == 'file':
            kwargs['filename'] = galaxy.modelpath
        elif data_origin == 'db':
            kwargs['plateifu'] = galaxy.plateifu
        elif data_origin == 'api':
            kwargs['plateifu'] = galaxy.plateifu
            kwargs['mode'] = 'remote'

        model_cube = ModelCube(**kwargs)
        assert model_cube.data_origin == data_origin
        self._test_init(model_cube, galaxy, bintype=galaxy.bintype.name)


class TestModelCube(object):

    @marvin_test_if(mark='include', galaxy={'plateifu': '8485-1901'})
    def test_get_flux_db(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu)
        shape = tuple([4563] + galaxy.shape)
        assert model_cube.binned_flux.shape == shape

    @marvin_test_if(mark='include', galaxy={'plateifu': '8485-1901'})
    def test_get_flux_remote(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu, mode='remote')
        shape = tuple([4563] + galaxy.shape)
        assert model_cube.binned_flux.shape == shape

    def test_get_cube_file(self, galaxy):
        model_cube = ModelCube(filename=galaxy.modelpath)
        assert isinstance(model_cube.getCube(), Cube)

    def test_get_cube_units(self, galaxy):
        model_cube = ModelCube(filename=galaxy.modelpath)
        unit = '1E-17 erg/s/cm^2/ang/spaxel'
        fileunit = model_cube.data['EMLINE'].header['BUNIT']
        assert unit == fileunit

        unit = fileunit.replace('ang', 'angstrom').split('/spaxel')[0]
        spaxel = u.Unit('spaxel', represents=u.pixel,
                        doc='A spectral pixel', parse_strict='silent')
        newunit = (u.Unit(unit) / spaxel)
        #unit = '1e-17 erg / (Angstrom cm2 s spaxel)'
        dmunit = model_cube.emline_fit.unit
        assert newunit == dmunit

    def test_get_maps_api(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu, mode='remote')
        assert isinstance(model_cube.getMaps(), Maps)

    def test_nobintype_in_db(self, galaxy):

        if galaxy.release != 'MPL-6':
            pytest.skip('only running this test for MPL6')

        with pytest.raises(MarvinError) as cm:
            ModelCube(plateifu=galaxy.plateifu, bintype='ALL', release=galaxy.release)

        assert 'Specified bintype ALL is not available in the DB' in str(cm.value)

    @pytest.mark.parametrize('objtype, errmsg',
                             [('cube', 'Trying to open a non DAP file with Marvin ModelCube'),
                              ('maps', 'Trying to open a DAP MAPS with Marvin ModelCube')])
    def test_modelcube_wrong_file(self, galaxy, objtype, errmsg):
        path = galaxy.cubepath if objtype == 'cube' else galaxy.mapspath
        with pytest.raises(MarvinError) as cm:
            ModelCube(filename=path)
        assert errmsg in str(cm.value)


class TestPickling(object):

    def test_pickling_file(self, temp_scratch, galaxy):
        modelcube = ModelCube(filename=galaxy.modelpath, bintype=galaxy.bintype)
        assert modelcube.data_origin == 'file'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is not None

        file = temp_scratch.join('test_modelcube.mpf')
        modelcube.save(str(file))

        assert file.check() is True
        assert modelcube.data is not None

        modelcube = None
        assert modelcube is None

        modelcube_restored = ModelCube.restore(str(file))
        assert modelcube_restored.data_origin == 'file'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is not None

    def test_pickling_file_custom_path(self, temp_scratch, galaxy):
        modelcube = ModelCube(filename=galaxy.modelpath, bintype=galaxy.bintype)
        assert modelcube.data_origin == 'file'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is not None

        file = temp_scratch.join('mcpickle').join('test_modelcube.mpf')
        assert file.check(file=1) is False

        path = modelcube.save(path=str(file))
        assert file.check() is True
        assert os.path.exists(path)

        modelcube_restored = ModelCube.restore(str(file), delete=True)
        assert modelcube_restored.data_origin == 'file'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is not None

        assert not os.path.exists(path)

    def test_pickling_db(self, galaxy, temp_scratch):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype)

        file = temp_scratch.join('test_modelcube_db.mpf')
        with pytest.raises(MarvinError) as cm:
            modelcube.save(str(file))

        assert 'objects with data_origin=\'db\' cannot be saved.' in str(cm.value)

    def test_pickling_api(self, temp_scratch, galaxy):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype, mode='remote')
        assert modelcube.data_origin == 'api'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is None

        file = temp_scratch.join('test_modelcube_api.mpf')
        modelcube.save(str(file))

        assert file.check() is True

        modelcube = None
        assert modelcube is None

        modelcube_restored = ModelCube.restore(str(file))
        assert modelcube_restored.data_origin == 'api'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is None


class TestMaskbit(object):

    def test_quality_flag(self, galaxy):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype)
        assert modelcube.quality_flag is not None

    @pytest.mark.parametrize('flag',
                             ['manga_target1',
                              'manga_target2',
                              'manga_target3',
                              'target_flags'])
    def test_flag(self, flag, galaxy):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype)
        assert getattr(modelcube, flag, None) is not None



from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

import distutils.spawn
import re
import time
from os import popen
from os.path import exists, dirname, join, basename, getsize, getmtime, sep
from datetime import datetime, timedelta
from sdss_access import AccessError
from sdss_access.sync.baseaccess import BaseAccess
from sdss_access import is_posix

try:
    from urllib2 import (HTTPPasswordMgrWithDefaultRealm, HTTPBasicAuthHandler, build_opener,
                         install_opener, urlopen)
except Exception:
    from urllib.request import (HTTPPasswordMgrWithDefaultRealm, HTTPBasicAuthHandler, build_opener,
                                install_opener, urlopen)


class CurlAccess(BaseAccess):
    """Class for providing Curl access to SDSS SAS Paths
    """
    remote_scheme = 'https'
    access_mode = 'curl'

    def __init__(self, label='sdss_curl', stream_count=5, mirror=False, public=False, release=None,
                 verbose=False):

        if not distutils.spawn.find_executable('curl'):
            msg = ('cURL does not appear to be installed. To install, the cURL '
                   'download wizard is located at: https://curl.haxx.se/dlwiz/. '
                   'Installation tutorials for cURL (software from https://curl.haxx.se) '
                   'are available online.')
            raise AccessError(msg)

        super(CurlAccess, self).__init__(stream_count=stream_count, mirror=mirror, public=public,
                                         release=release, verbose=verbose, label=label)

    def __repr__(self):
        return '<CurlAccess(using="{0}")>'.format(self.netloc)

    def get_task_status(self, task=None):
        if task:
            try:
                self.set_url_list(task['source'])
                is_there_any_files = len(self.file_line_list) > 0
                err = 'Found no files' if not is_there_any_files else ''
            except Exception as e:
                err = e
                is_there_any_files = False

            if not is_there_any_files:
                raise AccessError("Return code %r\n" % err)
        else:
            is_there_any_files = False
        return is_there_any_files

    def set_url_password(self, url_directory):
        """ Authorize User on sas"""
        url_directory = url_directory.split('sas')[0]
        password_mgr = HTTPPasswordMgrWithDefaultRealm()
        password_mgr.add_password(None, url_directory, self.auth.username, self.auth.password)
        handler = HTTPBasicAuthHandler(password_mgr)
        opener = build_opener(handler)
        opener.open(url_directory)
        install_opener(opener)

    def get_query_list(self, url_query):
        """Search through user specified "*" options and return all possible and valid url paths"""

        # Find query locations and set a descriptive dictionary
        query_objects = [{'segment_number': index, 'query_directory': '', 'query_list_index': 0,
                          'query_list': [], 'query': query.replace('*', '.*')} for index, query in
                         enumerate(url_query.split('/')) if '*' in query or index == len(url_query.split('/')) - 1]

        # Set quick use observables of query_objects
        segment_numbers = [query_object['segment_number'] for query_object in query_objects]
        max_depth = len(query_objects) - 1

        # Set url array used to append user specified urls
        query_results = []

        # Walk and search through optional branches for potential urls that pass user specifications
        query_depth = None

        if self.verbose:
            print("SDSS_ACCESS> Expanding wildcards %r" % url_query)

        while query_depth != 0:
            if query_depth is None:
                query_depth = 0

            # Set branch directory
            query_objects[query_depth]['query_directory'] = ''
            for segment_index, segment in enumerate(url_query.split('/')[:query_objects[query_depth]['segment_number']]):
                if segment_index not in segment_numbers:
                    query_objects[query_depth]['query_directory'] = '/'.join([query_objects[query_depth]['query_directory'], segment] if query_objects[query_depth]['query_directory'] else [segment])
                else:
                    query_object = query_objects[segment_numbers.index(segment_index)]
                    query_branch = query_object['query_list'][query_object['query_list_index']]
                    query_objects[query_depth]['query_directory'] = '/'.join([query_objects[query_depth]['query_directory'], query_branch])

            # Get user specified url options at branch directory
            try:
                query_objects[query_depth]['query_list'] = [item.split('"')[0] for item in re.findall(r'<a href="(%s)%s".*</a></td><td>'%(query_objects[query_depth]['query'], '/' if query_depth != max_depth else ''), urlopen(query_objects[query_depth]['query_directory']).read().decode('utf-8').replace('<tr><td><a href="../">Parent directory/</a></td><td>-</td><td>-</td></tr>', ''))]
            except Exception as e:
                query_objects[query_depth]['query_list'] = []
                if 'Unauthorized' in e:
                    raise AccessError("Return code %r\n" % e)

            # Append full url's that fit user specifications
            if query_depth == max_depth and len(query_objects[query_depth]['query_list']):
                for item in query_objects[query_depth]['query_list']:
                    query_results.append('/'.join([query_objects[query_depth]['query_directory'], item]))

            # Specify walker logic to recognize when to step down the branch or back up and go to the next option
            if not len(query_objects[query_depth]['query_list']) or query_depth == max_depth:
                query_depth -= 1
                while query_depth > -1 and query_objects[query_depth]['query_list_index'] == len(query_objects[query_depth]['query_list'])-1:
                    query_objects[query_depth]['query_list_index'] = 0
                    query_depth -= 1
                query_objects[query_depth]['query_list_index'] += 1
                query_depth += 1
            else:
                query_depth += 1

        return query_results

    def set_url_list(self, query_path=None):
        """Gets url paths from get_query_list and returns file proparties and path"""
        if not is_posix:
            query_path = query_path.replace(sep, '/')
        if not self.public:
            self.set_url_password(query_path)

        self.file_line_list, self.file_size_list, self.file_date_list, self.url_list = [], [], [], []
        for url in self.get_query_list(query_path):
            file_line, file_size, file_date = re.findall(r'<a href="(%s)".*</a></td><td>\s*(\d*)</td><td>(.*)</td></tr>\r' % basename(url), urlopen(dirname(url)).read().decode('utf-8'))[0]
            self.url_list.append(url)
            self.file_line_list.append(file_line.split('"')[0])
            self.file_size_list.append(file_size)
            self.file_date_list.append(file_date)

    def generate_stream_task(self, task=None, out=None):
        ''' creates the task to put in the download stream '''
        if task:
            location = task['location']
            for filename, file_size, file_date, url in zip(self.file_line_list, self.file_size_list, self.file_date_list, self.url_list):
                location = url.split('/sas/')[-1]
                source = join(self.stream.source, location) if self.remote_base else None
                destination = join(self.stream.destination, location)
                if not is_posix:
                    source = source.replace(sep, '/')
                    destination = destination.replace('/', sep)
                    location = location.replace('/', sep)
                if not self.check_file_exists_locally(destination, file_size, file_date):
                    yield (location, source, destination)

    def check_file_exists_locally(self, destination=None, url_file_size=None, url_file_time=None):
        """Checks if file already exists (note that time check is only accurate to the minute)"""
        if exists(destination):
            existing_file_size = int(popen('gzip -l %s' % destination).readlines()[1].split()[0]) if '.gz' in destination else getsize(destination)
            url_file_time = datetime.strptime(url_file_time, "%Y-%b-%d %H:%M" if len(url_file_time.split('-')[0]) == 4 else "%d-%b-%Y %H:%M")
            local_file_time = datetime.utcfromtimestamp(getmtime(destination))
            url_file_time = url_file_time + timedelta(seconds=time.altzone if time.daylight else time.timezone)
            if existing_file_size == int(url_file_size) and abs(url_file_time - local_file_time).seconds < 60:
                print('Already Downloaded at %s' % destination)
                return True
            else:
                return False
        else:
            return False

    def set_stream_task(self, task=None):
        status = self.get_task_status(task=task)
        if status:
            super(CurlAccess, self).set_stream_task(task=task)

    def _get_stream_command(self):
        ''' gets the stream command used when committing the download '''
        auth = ''
        if self.auth.username and self.auth.password:
            auth = '-u {0}:{1}'.format(self.auth.username, self.auth.password)
        return "curl {0} --create-dirs --fail -sSRLK {{path}}".format(auth)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-11-30 12:19:16
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-12-01 16:11:33

from __future__ import print_function, division, absolute_import
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils import statemachine
import traceback
import importlib
import re

# This file is a basic example of how to write a Sphinx plugin to document
# custom data.


def _indent(text, level=1):
    ''' Does a proper indenting for Sphinx rst '''

    prefix = ' ' * (4 * level)

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if line.strip() else line)

    return ''.join(prefixed_lines())


def _format_command(name, envvars, base=None):
    ''' Creates a list-table directive

    for a set of defined environment variables

    Parameters:
        name (str):
            The name of the config section
        envvars (dict):
            A dictionary of the environment variable definitions from the config
        base (str):
            The SAS_BASE to remove from the filepaths

    Yields:
        A string rst-formated list-table directive

    '''

    yield '.. list-table:: {0}'.format(name)
    yield _indent(':widths: 20 50')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Path')
    for envvar, path in envvars.items():
        tail = path.split(base)[1] if base and base in path else path
        tail = envvar.upper() if envvar.upper() == 'SAS_BASE_DIR' else tail
        yield _indent('* - {0}'.format(envvar.upper()))
        yield _indent('  - {0}'.format(tail))
    yield ''


def _format_changelog(changes):
    ''' Format a changelog for a Tree '''

    yield '.. role:: maroon'

    # check if changelog has a PATHS section
    has_paths = 'Changes in PATHS:' in changes
    if has_paths:
        path_idx = changes.index('Changes in PATHS:')
        path_changes = changes[path_idx:]
        changes = changes[:path_idx]

    # yield lines from the Environment section
    for line in changes:
        if 'Environment' in line or 'Changes' in line:
            yield f'**{line}**'
        elif re.search('[a-z]', line):
            new = re.sub(r'(New\s|Updated\s|^)(.*?):', r'\1:maroon:`\2`:', line)
            yield '* ' + new
        yield ''

    # yield lines from the PATHS section
    if has_paths:
        # list mode
        for line in path_changes:
            if 'Changes' in line:
                yield f'**{line}**'
                yield ''
            elif 'New' in line or 'Updated' in line:
                yield '* ' + _indent(f'**{line.strip()}**')
            elif re.search('[a-z]', line):
                # nested list From To
                if line.strip().startswith(('from', 'to')):
                    line = line.replace('to:', '**To:**').replace('from:', '**From:**')
                    yield _indent('    * ' + line, level=2)
                else:
                    name, template = line.split(':', 1)
                    yield _indent(f'    * :maroon:`{name}`: {template}')


def load_module(module_path, error=None, products=None):
    """Load the module."""

    # Exception to raise
    error = error if error else RuntimeError

    # __import__ will fail on unicode,
    # so we ensure module path is a string here.
    module_path = str(module_path)
    try:
        module_name, attr_name = module_path.split(':', 1)
    except ValueError:  # noqa
        raise error('"{0}" is not of format "module:object"'.format(module_path))

    # import the module
    try:
        mod = importlib.import_module(module_name)
    except (Exception, SystemExit) as exc:
        err_msg = 'Failed to import module "{0}". '.format(module_name)
        if isinstance(exc, SystemExit):
            err_msg += 'The module appeared to call sys.exit()'
        else:
            err_msg += 'The following exception was raised:\n{0}'.format(
                traceback.format_exc())

        raise error(err_msg)

    # check what kind of module
    if products:
        module = mod.dm.products
    else:
        module = mod

    if not hasattr(module, attr_name):
        raise error('Module "{0}" has no attribute "{1}"'.format(module_name, attr_name))

    return getattr(module, attr_name)


class TreeDirective(rst.Directive):
    ''' The directive which instructs Sphinx how to format the Tree config documentation '''

    has_content = False
    required_arguments = 1
    option_spec = {
        'prog': directives.unchanged_required,
        'title': directives.unchanged,
        'remove-sasbase': directives.flag,
    }

    def _generate_section(self, name, config, cfg_section='default', remove_sasbase=False):
        """Generate the relevant Sphinx nodes.

        Generates a section for the Tree datamodel.  Formats a tree section
        as a list-table directive.

        Parameters:
            name (str):
                The name of the config to be documented, e.g. 'sdsswork'
            config (dict):
                The tree dictionary of the loaded config environ
            cfg_section (str):
                The section of the config to load
            remove_sasbase (bool):
                If True, removes the SAS_BASE_DIR from the beginning of each path

        Returns:
            A section docutil node

        """

        # the source name
        source_name = name

        # Title
        section = nodes.section(
            '',
            nodes.title(text=cfg_section),
            ids=[nodes.make_id(cfg_section)],
            names=[nodes.fully_normalize_name(cfg_section)])

        # Summarize
        result = statemachine.ViewList()
        base = config['default']['FILESYSTEM'] if remove_sasbase else None
        lines = _format_command(cfg_section, config[cfg_section], base=base)
        for line in lines:
            result.append(line, source_name)

        self.state.nested_parse(result, 0, section)

        return [section]

    def run(self):
        self.env = self.state.document.settings.env

        command = load_module(self.arguments[0])

        if 'prog' in self.options:
            prog_name = self.options.get('prog')
        else:
            raise self.error(':prog: must be specified')

        # get instance and config
        instance = command(config=prog_name)
        config = instance.environ

        # get options
        remove_sasbase = 'remove-sasbase' in self.options

        # add a new section for each config part
        sections = []
        for cfg_section, envars in config.items():
            if cfg_section != 'default':
                sections.extend(self._generate_section(prog_name, config, cfg_section=cfg_section,
                                remove_sasbase=remove_sasbase))
        return sections


class TreeChangeDirective(rst.Directive):
    has_content = False
    required_arguments = 1
    option_spec = {
        'prog': directives.unchanged_required,
        'title': directives.unchanged,
        'drs': directives.unchanged_required,
        'remove-sasbase': directives.flag,
    }

    def run(self):
        self.env = self.state.document.settings.env

        command = load_module(self.arguments[0])

        if 'prog' in self.options:
            prog_name = self.options.get('prog')
        else:
            raise self.error(':prog: must be specified')

        # get options
        remove_sasbase = 'remove-sasbase' in self.options

        # get drs
        drs = self.options.get('drs')
        new, old = drs.replace(' ', '').split(',')

        # compute and format the changelog lines
        lines = command(new, old, remove_sas=remove_sasbase, to_list=True)
        title = lines[0]
        lines = lines[1:]
        lines = _format_changelog(lines)

        # the source name
        source_name = title

        # Title
        section = nodes.section(
            '',
            nodes.title(text=source_name),
            ids=[nodes.make_id(source_name)],
            names=[nodes.fully_normalize_name(source_name)])

        result = statemachine.ViewList()
        for line in lines:
            result.append(line, source_name)

        self.state.nested_parse(result, 0, section)

        return [section]


def setup(app):
    app.add_directive('datamodel', TreeDirective)
    app.add_directive('changelog', TreeChangeDirective)






# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: test_setup.py
# Project: tests
# Author: Brian Cherinka
# Created: Saturday, 13th April 2019 6:11:21 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Monday, 6th April 2020 6:46:44 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import os
import subprocess
import pytest
import glob
import sys
from tree import Tree

setuppath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../bin/setup_tree.py'))


@pytest.fixture()
def tree(faketree):
    # create the fake tree
    t = Tree()
    yield t
    t = None


def test_intemp(tree):
    assert 'treetmp' in tree.environ['general']['SAS_BASE_DIR']
    treedir = os.getenv('TREE_DIR')
    assert 'treetmp' in treedir
    assert 'treetmp' in os.getenv('MODULES_DIR')
    assert os.path.exists(os.path.join(treedir, 'etc'))
    assert os.path.exists(os.path.join(treedir, 'data'))
    files = glob.glob(os.path.join(treedir, 'data/*.cfg'))
    assert len(files) > 1


# def run_cmd(args=[], user_input=''):
#     process = subprocess.Popen(['python', setuppath] + args, universal_newlines=True,
#                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#     stdout, __ = process.communicate(input=user_input)
#     return stdout

def run_cmd(args=[], user_input=''):
    if sys.version_info.major == 2 or (sys.version_info.major == 3 and sys.version_info.minor < 7):
        out = subprocess.check_output(['python', setuppath] + args,
                                      universal_newlines=True, input=user_input)
    else:
        process = subprocess.run(['python', setuppath] + args, universal_newlines=True,
                                 input=user_input, capture_output=True)
        out = process.stdout
    return out


def read_index(path):
    with open(path, 'r') as f:
        data = f.read()
    return data


def assert_stuff(envdir, name):
    index = os.path.join(envdir, 'index.html')
    assert os.path.exists(index)
    page = read_index(index)
    assert name.upper() in page


@pytest.mark.parametrize('name', [('MANGA_HI'), ('MANGA_SPECTRO_REDUX')], ids=['mangahi', 'redux'])
def test_envlinks(tree, name):
    assert os.path.isfile(setuppath)
    envdir = os.path.join(tree.environ['general']['SAS_BASE_DIR'], 'dr15/env')
    stdout = run_cmd(args=['-e'])
    assert os.path.exists(envdir) is True
    assert 'Processing {0}'.format(name) in str(stdout)
    linkdir = os.path.join(envdir, name.upper())
    assert os.path.isdir(linkdir)
    assert os.path.islink(linkdir)
    assert_stuff(envdir, name)


@pytest.mark.parametrize('name', [('MANGA_HI'), ('MANGA_SPECTRO_REDUX')], ids=['mangahi', 'redux'])
def test_env_only(tree, name):
    dr15envdir = os.path.join(tree.environ['general']['SAS_BASE_DIR'], 'dr15/env')
    dr14envdir = os.path.join(tree.environ['general']['SAS_BASE_DIR'], 'dr14/env')
    stdout = run_cmd(args=['-e', '-o', 'dr14'])
    assert os.path.exists(dr14envdir) is True
    assert os.path.exists(dr15envdir) is False
    if 'HI' in name:
        assert 'Processing {0}'.format(name) not in str(stdout)
    else:
        assert 'Processing {0}'.format(name) in str(stdout)
        assert_stuff(dr14envdir, name)


@pytest.mark.parametrize('mirror, exp', [(False, 'SDSS-IV Science Archive Server (SAS)'),
                                         (True, 'SDSS-IV Science Archive Mirror (SAM)')])
def test_env_mirror(tree, mirror, exp):
    envdir = os.path.join(tree.environ['general']['SAS_BASE_DIR'], 'dr15/env')
    args = ['-e', '-i'] if mirror else ['-e']
    stdout = run_cmd(args=args)
    index = os.path.join(envdir, 'index.html')
    page = read_index(index)
    assert exp in page


@pytest.mark.parametrize('config', [('dr15'), ('dr14'), ('sdsswork'), ('sdss5')])
def test_create_files(tree, config):
    etcdir = os.path.join(os.getenv('TREE_DIR'), 'etc')
    assert os.path.exists(etcdir)
    files = glob.glob(os.path.join(etcdir, '*.*'))
    assert len(files) == 0
    stdout = run_cmd(args=[])
    files = glob.glob(os.path.join(etcdir, '*.*'))
    assert len(files) > 1
    assert os.path.exists(os.path.join(etcdir, config + '.sh'))
    assert os.path.exists(os.path.join(etcdir, config + '.csh'))
    assert os.path.exists(os.path.join(etcdir, config + '.module'))

    # ensure modules were copied
    moduledir = os.path.join(os.getenv('MODULES_DIR'), 'tree')
    files = glob.glob(os.path.join(moduledir, '*'))
    assert len(files) > 0
    assert os.path.exists(os.path.join(moduledir, config))


@pytest.fixture()
def resetmod(monkeypatch):
    mdir = os.environ.get('MODULES_DIR')
    base = os.path.dirname(mdir)
    monkeypatch.setenv("MODULES_DIR", '')
    os.makedirs(os.path.join(base, 'crap'))
    newpath = os.path.join(base, 'crap') + ':' + mdir
    monkeypatch.setenv('MODULEPATH', newpath)
    os.makedirs(os.path.join(mdir, 'tree'))


def assert_paths(path, config):
    files = glob.glob(os.path.join(path, '*'))
    assert os.path.exists(path)
    assert len(files) >= 0
    assert os.path.exists(os.path.join(path, config))


def assert_no_paths(path, config, haspath=False):
    files = glob.glob(os.path.join(path, '*'))
    assert os.path.exists(path) if haspath else not os.path.exists(path)
    assert len(files) == 0


@pytest.mark.parametrize('config, inputs',
                         [('dr15', '1'),
                          ('sdsswork', '2\ny\n'),
                          ('sdsswork', 'all\ny\n')], ids=['copy1', 'copy2', 'copyall'])
def test_emptymodulepath(tree, resetmod, config, inputs):

    modulepath = os.environ.get('MODULEPATH')
    split_mods = modulepath.split(':')
    for mpath in split_mods:
        if 'crap' in mpath:
            assert not os.path.exists(os.path.join(mpath, 'tree'))
        else:
            assert os.path.exists(os.path.join(mpath, 'tree'))

    stdout = run_cmd(args=[], user_input=inputs)

    if inputs == '1':
        # assert only first module path gets copied
        # crap path
        path = os.path.join(split_mods[0], 'tree')
        assert_paths(path, config)

        # module path
        path = os.path.join(split_mods[1], 'tree')
        assert_no_paths(path, config, haspath=True)

    elif inputs.startswith('2'):
        # assert only second module path gets copied
        # crap path
        path = os.path.join(split_mods[0], 'tree')
        assert_no_paths(path, config)

        # module path
        path = os.path.join(split_mods[1], 'tree')
        assert_paths(path, config)
    elif inputs.startswith('all'):
        # assert both module paths get copied
        for mpath in split_mods:
            path = os.path.join(mpath, 'tree')
            assert_paths(path, config)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Michael Talbot
# @Date:   2019-08-7 12:30:00
# @Last modified by:   Michael Talbot
# @Last Modified time: 2019-08-07 12:30:00

from __future__ import print_function, division, absolute_import
import os
import pytest


class TestCurl(object):

    def test_curl(self, curl, datapath):
        name = curl.release.lower() if curl.release and 'dr' in curl.release else datapath['work']
        path = curl.url(datapath['name'], **datapath['params'])
        assert datapath['location'] in path
        assert 'https://data.sdss.org' in path
        assert name in path

    # def test_get_locations(self, cstream, datapath):
    #     location = cstream.get_locations()[0]
    #     assert location == datapath['location']



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-03-24 12:22:07
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-08 15:47:42

from __future__ import print_function, division, absolute_import
import os
import pytest
from sdss_access import Access, AccessError
from sdss_access.sync import RsyncAccess


class TestRsync(object):

    def test_rsync(self, radd, expdata):
        assert radd.remote_base == expdata['base']
        assert radd.base_dir == os.environ.get("SAS_BASE_DIR") + '/'

    def test_get_locations(self, rstream, expdata):
        location = rstream.get_locations()[0]
        assert location == expdata['location']

    def test_get_paths(self, rstream, expdata):
        path = rstream.get_paths()[0]
        assert path == expdata['destination']

    @pytest.mark.slow
    def test_commit(self, monkeypatch, tmpdir, rsync):
        sasdir = tmpdir.mkdir("sas")
        monkeypatch.setenv("SAS_BASE_DIR", str(sasdir))

        # need to replant tree since rsync is a session level fixture
        rsync.replant_tree()
        rsync.add('drpall', drpver='v2_4_3')
        rsync.set_stream()
        rsync.commit()

        path = rsync.get_paths()[0]
        print('path', path)
        assert os.path.exists(path) is True
        assert os.path.isfile(path) is True

    @pytest.mark.parametrize('tree_ver, exp', [('sdsswork', 'work'), ('dr15', 'dr15'),
                                               ('dr13', 'dr13'), ('mpl8', 'work')])
    def test_release_from_module(self, monkeypatch, tree_ver, exp, datapath):
        monkeypatch.setenv('TREE_VER', tree_ver)
        rsync = RsyncAccess()
        rsync.remote()
        rsync.add(datapath['name'], **datapath['params'])
        loc = rsync.initial_stream.task[0]['location']
        assert rsync.release == tree_ver
        assert exp in loc


class TestRsyncFails(object):

    def test_access_svn_fail(self):
        with pytest.raises(AccessError) as cm:
            access = Access()
            access.remote()
            access.add('mangapreimg', designid=8405, designgrp='D0084XX', mangaid='1-42007')
        assert 'Rsync/Curl Access not allowed for svn paths.  Please use HttpAccess.' in str(cm.value)


class TestStream(object):
    def test_initial_stream(self, radd, inittask):
        task = radd.initial_stream.task
        assert task == inittask

    def test_final_stream(self, rstream, finaltask):
        task = rstream.stream.task
        assert task == finaltask



from .auth import Auth, AuthMixin
from .cli import Cli
from .stream import Stream
from .http import HttpAccess
from .baseaccess import BaseAccess
from .rsync import RsyncAccess
from .curl import CurlAccess
from .access import Access


from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

import abc
import six
from os.path import join, sep
from sdss_access import SDSSPath
from sdss_access.sync.auth import Auth, AuthMixin
from sdss_access.sync.stream import Stream
from sdss_access import is_posix, AccessError


class BaseAccess(six.with_metaclass(abc.ABCMeta, AuthMixin, SDSSPath)):
    """Class for providing Rsync or Curl access to SDSS SAS Paths
    """
    remote_scheme = None
    access_mode = 'rsync' if is_posix else 'curl'

    def __init__(self, label=None, stream_count=5, mirror=False, public=False, release=None,
                 verbose=False, force_modules=None):
        super(BaseAccess, self).__init__(release=release, public=public,
                                         mirror=mirror, verbose=verbose,
                                         force_modules=force_modules)
        self.label = label
        self.auth = None
        self.stream = None
        self.stream_count = stream_count
        self._stream_command = None
        self.verbose = verbose
        self.initial_stream = self.get_stream()

    def remote(self, username=None, password=None, inquire=None):
        """ Configures remote access """
        use_dtn = self.remote_scheme == 'rsync' and not self.public
        # simplifies things to have a single sdss machine in .netrc
        self.set_netloc(sdss=True)
        self.set_auth(username=username, password=password, inquire=inquire)
        if use_dtn:
            self.set_netloc(dtn=use_dtn)
        self.set_remote_base(scheme=self.remote_scheme)

    def add(self, filetype, **kwargs):
        """ Adds a filepath into the list of tasks to download"""

        location = self.location(filetype, **kwargs)

        # set proper sasdir based on access method
        if self.access_mode == 'rsync':
            sasdir = 'sas' if not self.public else ''
        elif self.access_mode == 'curl':
            sasdir = 'sas'
        source = self.url(filetype, sasdir=sasdir, **kwargs)

        # raise error if attempting to add a software product path
        if 'svn.sdss.org' in source:
            raise AccessError('Rsync/Curl Access not allowed for svn paths.  Please use HttpAccess.')

        if 'full' not in kwargs:
            destination = self.full(filetype, **kwargs)
        else:
            destination = kwargs.get('full')

        if location and source and destination:
            self.initial_stream.append_task(
                location=location, source=source, destination=destination)
        else:
            print("There is no file with filetype=%r to access in the tree module loaded" % filetype)

    def set_stream(self):
        """ Sets the download streams """

        if not self.auth:
            raise AccessError(
                "Please use the remote() method to set rsync authorization or use remote(public=True) for public data")
        elif not self.initial_stream.task:
            raise AccessError("No files to download.")
        else:
            self.stream = self.get_stream()

            # set stream source based on access mode
            if self.access_mode == 'rsync':
                source = join(self.remote_base, self.release.lower()
                              if self.public else 'sas') if self.remote_base else self.remote_base
            elif self.access_mode == 'curl':
                source = join(self.remote_base, 'sas').replace(sep, '/')
            self.stream.source = source

            # set stream destination based on access mode
            if self.access_mode == 'rsync':
                dest = join(self.base_dir, self.release.lower()
                            if self.public else '') if self.base_dir else self.base_dir
            elif self.access_mode == 'curl':
                dest = self.base_dir
            self.stream.destination = dest

            # set client env dict based on access mode
            if self.access_mode == 'rsync':
                key = 'RSYNC_PASSWORD'
            elif self.access_mode == 'curl':
                key = 'CURL_PASSWORD'
            self.stream.cli.env = {key: self.auth.password} if self.auth.ready() else None

            if self.stream.source and self.stream.destination:
                for task in self.initial_stream.task:
                    self.set_stream_task(task)
            ntask = len(self.stream.task)
            if self.stream.stream_count > ntask:
                if self.verbose:
                    print("SDSS_ACCESS> Reducing the number of streams from %r to %r, the number of download tasks." % (
                        self.stream.stream_count, ntask))
                self.stream.stream_count = ntask
                self.stream.streamlet = self.stream.streamlet[:ntask]

    def get_stream(self):
        ''' return a Stream object '''
        stream = Stream(stream_count=self.stream_count, verbose=self.verbose)
        return stream

    def reset(self):
        ''' Reset all streams '''

        # reset the main stream
        if self.stream:
            self.stream.reset()

        # reset the initial stream (otherwise old 'adds' remain in the new stream)
        if self.initial_stream:
            self.initial_stream.reset()

    def shuffle(self):
        ''' Shuffle the stream '''
        self.stream.shuffle()

    def get_locations(self, offset=None, limit=None):
        ''' Rreturn the locations for all paths in the stream '''
        return self.stream.get_locations(offset=offset, limit=limit) if self.stream else None

    def get_paths(self, offset=None, limit=None):
        ''' Return the base paths for all paths in the stream '''
        locations = self.get_locations(offset=offset, limit=limit)
        sasdir = self.release.lower() if self.public else ''
        paths = [join(self.base_dir, sasdir, location) for location in locations] if locations else None
        return paths

    def get_urls(self, offset=None, limit=None):
        ''' Return the urls for all paths in the stream '''
        locations = self.get_locations(offset=offset, limit=limit)
        remote_base = self.get_remote_base()
        sasdir = self.release.lower() if self.public else 'sas'
        urls = [join(remote_base, sasdir, location) for location in locations] if locations else None
        return urls

    @abc.abstractmethod
    def generate_stream_task(self, task=None, out=None):
        ''' creates the task to put in the download stream '''

    def set_stream_task(self, task=None, out=None):
        ''' sets the path input dictionary for a task in a stream '''
        stream_has_task = False
        for location, source, destination in self.generate_stream_task(task=task, out=out):
            if location and source and destination:
                stream_has_task = True
                self.stream.append_task(location=location, source=source, destination=destination)
                """if self.verbose:
                    print("SDSS_ACCESS> Preparing to download: %s" % location)
                    print("SDSS_ACCESS> from: %s" % source)
                    print("SDSS_ACCESS> to: %s" % destination)
                    print("-"*80)"""

        if not stream_has_task:
            print('SDSS_ACCESS> Error: stream has nothing to do.')

    @abc.abstractmethod
    def _get_stream_command(self):
        ''' gets the stream command used when committing the download '''

    def commit(self, offset=None, limit=None):
        """ Start the download """

        self.stream.command = self._get_stream_command()
        self.stream.append_tasks_to_streamlets(offset=offset, limit=limit)
        self.stream.commit_streamlets()
        self.stream.run_streamlets()
        self.stream.reset_streamlet()



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-03-24 12:22:30
# @Last modified by:   Michael Talbot
# @Last Modified time: 2019-08-07 12:30:00

from __future__ import print_function, division, absolute_import
import glob
import gzip
import os
import pytest
import yaml
import contextlib
import shutil

from sdss_access import RsyncAccess, HttpAccess, CurlAccess
from sdss_access.path import Path


# PYTEST MODIFIERS
# -----------------
def pytest_addoption(parser):
    """Add new options"""
    # run slow tests
    parser.addoption('--runslow', action='store_true', default=False, help='Run slow tests.')


def pytest_runtest_setup(item):
    """Skip slow tests."""
    if 'slow' in item.keywords and not item.config.getoption('--runslow'):
        pytest.skip('Requires --runslow option to run.')


@pytest.fixture()
def path():
    ''' Fixture to create a generic Path object '''
    path = Path()
    path.replant_tree()
    yield path
    path = None


# releases to parametrize over
releases = ['work', 'DR15']


# read in test data parameters and also get paths
def get_data():
    ''' Retrieves the test data from the paths.yaml file '''

    with open(os.path.join(os.path.dirname(__file__), 'data/paths.yaml')) as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
    return data

data = get_data()
paths = data.get('paths')


@pytest.fixture(scope='session', params=releases)
def release(request):
    ''' release fixture '''
    return request.param


@pytest.fixture(scope='session', params=paths)
def datapath(request):
    ''' parametrizes over the paths in test data'''
    return request.param


@pytest.fixture(scope='session')
def expdata(release, datapath):
    ''' fixture to yield expected source data based on test data '''
    remote = data.get('remote_base')
    # remote base
    base = remote['work'] if release == 'work' else remote['public']
    # work or DR directory
    name = datapath['work'] if release == 'work' else ''
    # file location
    location = os.path.join(name, datapath['location'])
    # full source file location
    source = os.path.join(base, 'sas' if release == 'work' else release.lower(), location)
    # full final file location
    destination = os.path.join(os.getenv('SAS_BASE_DIR'), '' if release == 'work' else release.lower(), location)
    # combined dict
    result = {'name': datapath['name'], 'params': datapath['params'], 'base': base,
              'location': location, 'source': source, 'destination': destination, 'release': release.lower()}
    yield result
    result = None


@pytest.fixture(scope='session')
def inittask(expdata):
    ''' fixture to yield expected initial stream task based on test data '''

    patch = '' if expdata['release'] == 'work' else expdata['release']
    loc = os.path.join(patch, expdata['location'])
    task = [{'location': loc, 'source': expdata['source'],
             'destination': expdata['destination'], 'exists': None}]
    yield task
    task = None


@pytest.fixture(scope='session')
def finaltask(expdata):
    ''' fixture to yield expected final stream task based on test data '''

    task = [{'location': expdata['location'], 'source': expdata['source'],
             'destination': expdata['destination'], 'exists': None}]
    yield task
    task = None


@pytest.fixture(scope='session')
def rsync(release):
    ''' fixture to create generic rsync object - parametrized by release '''

    if 'DR' in release:
        rsync = RsyncAccess(label='test_rsync', public=True, release=release)
    else:
        rsync = RsyncAccess(label='test_rsync')
    rsync.remote()
    yield rsync
    # teardown
    rsync.reset()
    rsync = None


@pytest.fixture(scope='session')
def radd(rsync, expdata):
    ''' fixture to add a path to an rsync object '''
    rsync.add(expdata['name'], **expdata['params'])
    yield rsync
    rsync.reset()


@pytest.fixture(scope='session')
def rstream(radd):
    ''' fixture to set the stream for an parametrized rsync object '''
    radd.set_stream()
    yield radd
    radd.reset()


@pytest.fixture(scope='session')
def http(release):
    if 'DR' in release:
        http = HttpAccess(public=True, release=release)
    else:
        http = HttpAccess()
    yield http
    http = None


@pytest.fixture(scope='session')
def curl(release):
    ''' fixture to create generic curl object - parametrized by release '''

    if 'DR' in release:
        curl = CurlAccess(label='test_curl', public=True, release=release)
    else:
        curl = CurlAccess(label='test_curl')
    curl.remote()
    yield curl
    # teardown
    curl.reset()
    curl = None


@pytest.fixture(scope='session')
def cadd(curl, expdata):
    ''' fixture to add a path to an curl object '''
    curl.add(expdata['name'], **expdata['params'])
    yield curl


@pytest.fixture(scope='session')
def cstream(cadd):
    ''' fixture to set the stream for an parametrized curl object '''
    cadd.set_stream()
    yield cadd


@pytest.fixture()
def monkeysas(tmpdir, monkeypatch, path):
    ''' fixture to monkeypatch the sas_base_dir '''
    orig = os.getenv("SAS_BASE_DIR")
    tmppath = tmpdir / 'sas'
    os.makedirs(tmppath, exist_ok=True)
    monkeypatch.setenv('SAS_BASE_DIR', str(tmppath))
    path.replant_tree()
    yield
    os.environ["SAS_BASE_DIR"] = orig
    path.replant_tree()


@pytest.fixture()
def copydata(tmpdir, request):
    ''' fixture to copy a file into a temporary directory '''
    srcpath = os.path.join(os.getenv("SAS_BASE_DIR"), request.param)
    # skip the test if no real data exists to copy
    if not os.path.exists(srcpath):
        pytest.skip('file does not exist cannot copy')
    sasdir = tmpdir / 'sas'
    destpath = sasdir / request.param
    os.makedirs(os.path.dirname(destpath), exist_ok=True)
    shutil.copy(srcpath, destpath)
    yield destpath


@pytest.fixture()
def copymulti(tmpdir, request):
    ''' Fixture to copy multiple files into a temporary directory '''
    srcpath = os.path.join(os.getenv("SAS_BASE_DIR"), request.param)
    files = glob.glob(srcpath)
    if not files:
        pytest.skip('Files do not exist, cannot copy')
    for item in files:
        loc = item.split(os.getenv("SAS_BASE_DIR") + '/')[-1]
        sasdir = tmpdir / 'sas'
        destpath = sasdir / loc
        os.makedirs(os.path.dirname(destpath), exist_ok=True)
        shutil.copy(item, destpath)


@contextlib.contextmanager
def gzuncompress(filename):
    ''' Context manager than gunzips a file temporarily. '''
    import pathlib
    pp = pathlib.Path(filename)
    decompfile = pp.parent / pp.stem
    with gzip.open(filename, 'rb') as f_in:
        with open(decompfile, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(filename)
    yield


@contextlib.contextmanager
def gzcompress(filename):
    ''' Context manager than gzips a file temporarily. '''
    compfile = filename + '.gz'
    with open(filename, 'rb') as f_in:
        with gzip.open(compfile, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(filename)
    yield


@pytest.fixture()
def monkeyhome(monkeypatch, tmp_path):
    ''' monkeypatch the HOME directory '''
    path = (tmp_path / 'tmp').mkdir()
    monkeypatch.setenv("HOME", str(path))



from __future__ import division, print_function

import os
import re
import requests
import ast
import inspect
import six
from glob import glob
from os.path import join, sep
from random import choice, sample
from sdss_access import tree, log, config
from sdss_access import is_posix

pathlib = None
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

try:
    from ConfigParser import RawConfigParser
except ImportError:
    from configparser import RawConfigParser

"""
Module for constructing paths to SDSS files.

Example use case:

    from sdss_access.path import Path
    sdss_path = Path()
    filename = sdss_path.full('photoObj', run=94, rerun='301', camcol=1, field=100)

Depends on the tree product. In particular requires path templates in:
  $TREE_DIR/data/sdss_paths.ini
"""


class BasePath(object):
    """Class for construction of paths in general.

    Parameters
    ----------
    release : str
        The release name, e.g. 'DR15', 'MPL-9'.
    public : bool
        If True, uses public urls.  Only needed for public data releases. Automatically set to True when release contains "DR".
    mirror : bool
        If True, uses the mirror data domain url.  Default is False.
    verbose: bool
        If True, turns on verbosity.  Default is False.
    force_modules: bool
        If True, forces software products to use any existing Module environment paths

    Attributes
    ----------
    templates : dict
        The set of templates read from the configuration file.
    """

    _netloc = {"dtn": "sdss@dtn01.sdss.org", "sdss": "data.sdss.org",
               "mirror": "data.mirror.sdss.org", 'svn': 'svn.sdss.org'}

    def __init__(self, release=None, public=False, mirror=False, verbose=False, force_modules=None):
        self.release = release or os.getenv('TREE_VER', 'sdsswork')
        self.public = 'dr' in self.release.lower() or public
        self.mirror = mirror
        self.verbose = verbose
        self.force_modules = force_modules or config.get('force_modules')
        self.set_netloc()
        self.set_remote_base()

        self._compressions = ['.gz', '.bz2', '.zip']
        self._comp_regex = r'({0})$'.format('|'.join(self._compressions))
        # set the path templates from the tree
        self.templates = tree.paths
        if self.release:
            self.replant_tree(release=self.release)

    def replant_tree(self, release=None):
        ''' replants the tree based on release

        Resets the path definitions given a specified release

        Parameters:
            release (str):
                A release to use when replanting the tree
        '''
        release = release or self.release
        if release:
            release = release.lower().replace('-', '')
        tree.replant_tree(release)
        self.templates = tree.paths
        self.release = release

    @staticmethod
    def get_available_releases(public=None):
        ''' Get the available releases

        Parameters:
            public (bool):
                If True, only return public data releases
        '''
        return tree.get_available_releases(public=public)

    def lookup_keys(self, name):
        ''' Lookup the keyword arguments needed for a given path name

        Parameters:
            name (str):
                The name of the path

        Returns:
            A list of keywords needed for filepath generation

        '''

        assert name, 'Must specify a path name'
        assert name in self.templates.keys(), '{0} must be defined in the path templates'.format(name)
        # find all words inside brackets
        keys = list(set(re.findall(r'{(.*?)}', self.templates[name])))
        # lookup any keys referenced inside special functions
        skeys = self._check_special_kwargs(name)
        keys.extend(skeys)
        # remove any duplicates
        keys = list(set(keys))
        # remove the type : descriptor
        keys = [k.split(':')[0] for k in keys]
        return keys

    def _check_special_kwargs(self, name):
        ''' check special functions for kwargs

        Checks the content of the special functions (%methodname) for
        any keyword arguments referenced within

        Parameters:
            name (str):
                A path key name

        Returns:
            A list of keyword arguments found in any special functions
        '''
        keys = []
        # find any %method names in the template string
        functions = re.findall(r"\@\w+", self.templates[name])
        if not functions:
            return keys

        # loop over special method names and extract keywords
        for function in functions:
            method = getattr(self, function[1:])
            # get source code of special method
            source = self._find_source(method)
            fkeys = re.findall(r'kwargs\[(.*?)\]', source)
            if fkeys:
                # evaluate to proper string
                fkeys = [ast.literal_eval(k) for k in fkeys]
                keys.extend(fkeys)
        return keys

    @staticmethod
    def _find_source(method):
        ''' find source code of a given method

        Find and extract the source code of a given method in a module.
        Uses inspect.findsource to get all source code and performs some
        selection magic to identify method source code.  Doing it this way
        because inspect.getsource returns wrong method.

        Parameters:
            method (obj):
                A method object

        Returns:
            A string containing the source code of a given method

        Example:
            >>> from sdss_access.path import Path
            >>> path = Path()
            >>> path._find_source(path.full)
        '''

        # get source code lines of entire module method is in
        source = inspect.findsource(method)
        is_method = inspect.ismethod(method)
        # create single source code string
        source_str = '\n'.join(source[0])
        # define search pattern
        if is_method:
            pattern = r'def\s{0}\(self'.format(method.__name__)
        # search for pattern within the string
        start = re.search(pattern, source_str)
        if start:
            # find start and end positions of source code
            startpos = start.start()
            endpos = source_str.find('def ', startpos + 1)
            code = source_str[startpos:endpos]
        else:
            code = None
        return code

    def lookup_names(self):
        ''' Lookup what path names are available

        Returns a list of the available path names in sdss_access.
        Use with lookup_keys to find the required keyword arguments for a
        given path name.

        Returns:
            A list of the available path names.
        '''
        return self.templates.keys()

    def has_name(self, name):
        ''' Check if a given path name exists in the set of templates

        Parameters:
            name (str):
                The path name to lookup
        '''
        assert isinstance(name, six.string_types), 'name must be a string'
        return name in self.lookup_names()

    def extract(self, name, example):
        ''' Extract keywords from an example path

        Attempts to extract the defined keyword values from an example filepath
        for a given path name.  The filepath must be a full SDSS SAS filepath.

        Parameters:
            name (str):
                The name of the path definition
            example (str):
                The absolute filepath to the example file

        Returns:
            A dictionary of path keyword values

        Example:
            >>> from sdss_access.path import Path
            >>> path = Path()
            >>> filepath = '/Users/Brian/Work/sdss/sas/mangawork/manga/spectro/redux/v2_5_3/8485/stack/manga-8485-1901-LOGCUBE.fits'
            >>> path.extract('mangacube', filepath)
            >>> {'drpver': 'v2_5_3', 'plate': '8485', 'ifu': '1901', 'wave': 'LOG'}
        '''

        # if pathlib not available do nothing
        if not pathlib:
            return None

        # ensure example is a string
        if isinstance(example, pathlib.Path):
            example = str(example)
        assert isinstance(example, six.string_types), 'example file must be a string'

        # get the template
        assert name in self.lookup_names(), '{0} must be a valid template name'.format(name)
        template = self.templates[name]

        # expand the environment variable
        template = _expandvars(template)

        # handle special functions; perform a drop in replacement
        if re.match('@spectrodir', template):
            template = re.sub('@spectrodir', os.environ['BOSS_SPECTRO_REDUX'], template)
        elif re.search('@platedir', template):
            template = re.sub('@platedir', r'(.*)/{plateid:0>6}', template)
        elif re.search('@definitiondir', template):
            template = re.sub('@definitiondir', '{designid:0>6}', template)
        if re.search('@plateid6', template):
            template = re.sub('@plateid6', '{plateid:0>6}', template)

        # check if template has any brackets
        haskwargs = re.search('[{}]', template)
        if not haskwargs:
            return None

        # escape the envvar $ and any dots (use re in case of @platedir sub)
        template = self._remove_compression(template)
        subtemp = template.replace('$', '\\$')
        subtemp = re.sub(r'[.](?!\*)', '\\.', subtemp)
        # define search pattern; replace all template keywords with regex "(.*)" group
        research = re.sub('{(.*?)}', '(.*?)', subtemp)
        research += '$'  # mark the end of a search string (captures cases when {} at end of string)
        # look for matches in template and example
        pmatch = re.search(research, self._remove_compression(template))
        tmatch = re.search(research, self._remove_compression(example))

        path_dict = {}
        # if example match extract keys and values from the match groups
        if tmatch:
            values = tmatch.groups(0)
            keys = pmatch.groups(0)
            assert len(keys) == len(values), 'pattern and template matches must have same length'
            parts = zip(keys, values)
            # parse into dictionary
            for part in parts:
                value = part[1]
                if re.findall('{(.*?)}', part[0]):
                    # get the key name inside the brackets
                    keys = re.findall('{(.*?)}', part[0])
                    # remove the type : designation
                    keys = [k.split(':')[0] for k in keys]
                    # handle double bracket edge cases; remove this when better solution found
                    if len(keys) > 1:
                        if keys[0] == 'dr':
                            # for {dr}{version}
                            drval = re.match('^DR[1-9][0-9]', value).group(0)
                            otherval = value.split(drval)[-1]
                            pdict = {keys[0]: drval, keys[1]: otherval}
                        elif keys[0] in ['rc', 'br', 'filter', 'camrow']:
                            # for {camrow}{camcol}, {filter}{camcol}, {br}{id}, etc
                            pdict = {keys[0]: value[0], keys[1]: value[1:]}
                        else:
                            raise ValueError('This case has not yet been accounted for.')
                        path_dict.update(pdict)
                    else:
                        path_dict[keys[0]] = value
        return path_dict

    def dir(self, filetype, **kwargs):
        """Return the directory containing a file of a given type.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        dir : str
            Directory containing the file.
        """

        full = kwargs.get('full', None)
        if not full:
            full = self.full(filetype, **kwargs)

        return os.path.dirname(full)

    def name(self, filetype, **kwargs):
        """Return the name of a file of a given type.

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        name : str
            Name of a file with no directory information.
        """

        full = kwargs.get('full', None)
        if not full:
            full = self.full(filetype, **kwargs)

        return os.path.basename(full)

    def exists(self, filetype, remote=None, **kwargs):
        '''Checks if the given type of file exists locally

        Parameters
        ----------
        filetype : str
            File type parameter.

        remote : bool
            If True, checks for remote existence of the file

        Returns
        -------
        exists : bool
            Boolean indicating if the file exists.

        '''

        full = kwargs.get('full', None)
        if not full:
            full = self.full(filetype, **kwargs)

        if remote:
            # check for remote existence using a HEAD request
            url = self.url('', full=full)
            try:
                resp = requests.head(url, allow_redirects=True)
            except Exception as e:
                raise AccessError('Cannot check for remote file existence for {0}: {1}'.format(url, e))
            else:
                return resp.ok
        else:
            return os.path.isfile(full)

    def expand(self, filetype, **kwargs):
        ''' Expand a wildcard path locally

        Parameters
        ----------
        filetype : str
            File type parameter.

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        expand : list
            List of expanded full paths of the given type.

        '''

        full = kwargs.get('full', None)
        if not full:
            full = self.full(filetype, **kwargs)

        # assert '*' in full, 'Wildcard must be present in full path'
        files = glob(self._add_compression_wild(full))

        # return as urls?
        as_url = kwargs.get('as_url', None)
        newfiles = [self.url('', full=full) for full in files] if as_url else files

        # optionally refine the results
        refine = kwargs.get('refine', None)
        if refine:
            newfiles = self.refine(newfiles, refine, **kwargs)

        return newfiles

    def any(self, filetype, **kwargs):
        ''' Checks if the local directory contains any of the type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        Returns
        -------
        any : bool
            Boolean indicating if the any files exist in the expanded path on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        return any(expanded_files)

    def one(self, filetype, **kwargs):
        ''' Returns random one of the given type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        one : str
            Random file selected from the expanded list of full paths on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        isany = self.any(filetype, **kwargs)
        return choice(expanded_files) if isany else None

    def random(self, filetype, **kwargs):
        ''' Returns random number of the given type of file

        Parameters
        ----------
        filetype : str
            File type parameter.

        num : int
            The number of files to return

        as_url: bool
            Boolean to return SAS urls

        refine: str
            Regular expression string to filter the list of files by
            before random selection

        Returns
        -------
        random : list
            Random file selected from the expanded list of full paths on disk.

        '''
        expanded_files = self.expand(filetype, **kwargs)
        isany = self.any(filetype, **kwargs)
        if isany:
            # get the desired number
            num = kwargs.get('num', 1)
            assert num <= len(expanded_files), 'Requested number must be larger the sample.  Reduce your number.'
            return sample(expanded_files, num)
        else:
            return None

    def refine(self, filelist, regex, filterdir='out', **kwargs):
        ''' Returns a list of files filterd by a regular expression

        Parameters
        ----------
        filelist : list
            A list of files to filter on.

        regex : str
            The regular expression string to filter your list

        filterdir: {'in', 'out'}
            Indicates the filter to be inclusive or exclusive
            'out' removes the items satisfying the regular expression
            'in' keeps the items satisfying the regular expression

        Returns
        -------
        refine : list
            A file list refined by an input regular expression.

        '''
        assert filelist, 'Must provide a list of filenames to refine on'
        assert regex, 'Must provide a regular expression to refine the file list'
        r = re.compile(regex)

        # icheck filter direction; default is out
        assert filterdir in ['in', 'out'], 'Filter direction must be either "in" or "out"'
        if filterdir == 'out':
            subset = list(filter(lambda i: r.search(i), filelist))
        elif filterdir == 'in':
            subset = list(filter(lambda i: not r.search(i), filelist))
        return subset

    def full(self, filetype, **kwargs):
        """Return the full local path of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.
        force_module: bool
            If True, forces software products to use any existing Module environment paths
        kwargs: dict
            Any path template keyword arguments

        Returns
        -------
        full : str
            The full local path to the file.
        """

        # check if full already in kwargs
        if 'full' in kwargs:
            return kwargs.get('full')

        # check for filetype in template
        assert filetype in self.templates, ('No entry {0} found. Filetype must '
                                            'be one of the designated templates '
                                            'in the currently loaded tree'.format(filetype))
        template = self.templates[filetype]
        if not is_posix:
            template = template.replace('/', sep)

        # Check if forcing module paths
        force_module = kwargs.get('force_module', None)
        if force_module or self.force_modules:
            template = self.check_modules(template, permanent=self.force_modules)

        # Now replace {} items
        # check for missing keyword arguments
        keys = self.lookup_keys(filetype)
        # split keys to remove :format from any "key:format"
        keys = [k.split(':')[0] for k in keys]
        missing_keys = set(keys) - set(kwargs.keys())
        if missing_keys:
            raise KeyError('Missing required keyword arguments: {0}'.format(list(missing_keys)))
        else:
            template = template.format(**kwargs)

        # Now replace environmental variables
        template = _expandvars(template)

        # Now call special functions as appropriate
        template = self._call_special_functions(filetype, template, **kwargs)

        # Now match on any software product tags
        skip_tag_check = kwargs.get('skip_tag_check', None)
        if not skip_tag_check:
            template = re.sub(r'tags/(v?[0-9._]+)', r'\1', template, count=1)

        return self._check_compression(template)

    @staticmethod
    def check_modules(template, permanent=None):
        ''' Check for any existing Module path environment

        For software product paths, overrides the tree environment paths with existing
        original envvars from os.environ that may be set from shell bash or module environments.
        Checks the original os.environ for any environment variables and replaces the template
        envvar with the original os version.  Ignores all SAS data paths.  If permanent is True,
        then permanently replaces the envvar in existing os.environ with the original.
        Assumes original environment variables points to definitions created by module files or bash
        profiles.

        Parameters:
            template (str):
                The path template to check
            permanent (bool):
                If True, sets the original module environment variable into os.environ

        Returns:
            The template with updated environment variable path
        '''
        # if template starts with $SAS_BASE_DIR, then do nothing
        expanded_template = _expandvars(template)
        if expanded_template.startswith(os.getenv("SAS_BASE_DIR")):
            return template

        # match template against envvar $ENVVAR_DIR
        ev_match = re.match(r'^\$(\w+)', template)
        if ev_match:
            envvar = ev_match.group()
            envvar_name = ev_match.groups()[0]
            orig_os = tree.get_orig_os_environ()
            if envvar_name in orig_os:
                orig_envvar = orig_os.get(envvar_name)
                # update the real os environment
                if permanent:
                    os.environ[envvar_name] = orig_envvar
                return template.replace(envvar, orig_envvar)
            else:
                log.info('No existing envvar found for {0}. Returning input template'.format(envvar_name))
                return template

    def _remove_compression(self, template):
        ''' remove a compression suffix '''
        is_comp = re.search(self._comp_regex, template)
        if is_comp:
            temp_split = re.split(self._comp_regex, template)
            template = temp_split[0]
        return template

    def _add_compression_wild(self, template):
        ''' add a compression wildcard '''
        is_comp = re.search(self._comp_regex, template)
        if is_comp:
            for comp in self._compressions:
                template = template.replace(comp, '*')
        else:
            template = template + '*'
        return template

    def _check_compression(self, template):
        ''' check if filepath is actually compressed '''

        exists = self.exists('', full=template)
        if exists:
            return template

        # check if file is not compressed compared to template
        is_comp = re.search(self._comp_regex, template)
        if is_comp:
            base = os.path.splitext(template)[0]
            exists = self.exists('', full=base)
            if exists:
                return base

        # check if file on disk is actually compressed compared to template
        alternates = glob(template + '*')
        if alternates:
            suffixes = list(set([re.search(self._comp_regex, c).group(0)
                                 for c in alternates if re.search(self._comp_regex, c)]))
            if suffixes:
                assert len(suffixes) == 1, 'should only be one suffix per file template '
                if not template.endswith(suffixes[0]):
                    template = template + suffixes[0]

        return template

    def _call_special_functions(self, filetype, template, **kwargs):
        ''' Call the special functions found in a template path

        Calls special functions indicated by %methodname found in the
        sdss_paths.ini template file, and replaces the %location in the path
        with the returned content.

        Parameters:
            filetype (str):
                template name of file
            template (str):
                the template path
            kwargs (dict):
                Any kwargs needed to pass into the methods

        Returns:
            The expanded template path
        '''
        # Now call special functions as appropriate
        functions = re.findall(r"\@\w+", template)
        if not functions:
            return template

        for function in functions:
            try:
                method = getattr(self, function[1:])
            except AttributeError:
                return None
            else:
                value = method(filetype, **kwargs)
                template = template.replace(function, value)

        return template

    def get_netloc(self, netloc=None, sdss=None, dtn=None, svn=None, mirror=None):
        ''' Get a net url domain

        Returns an SDSS url domain location.  Options are the SDSS SAS domain, the rsync download
        server, the svn server, or the mirror data domain.  The mirror data domain is retrieved
        either by the ``mirror`` input keyword argument or by the ``path.mirror`` attribute.

        Parameters
        ----------
            netloc : str
                An exact net location to return directly
            sdss : bool
                If True, returns SDSS data domain: data.sdss.org
            dtn : bool
                If True, returns SDSS rsync server domain: dtn01.sdss.org
            svn: bool
                If True, returns SDSS svn domain: svn.sdss.org
            mirror: bool
                If True, return SDSS mirror domain: data.mirror.sdss.org.

        Returns
        -------
            An http domain name
        '''
        if netloc:
            return netloc

        if dtn:
            return self._netloc["dtn"]
        elif sdss:
            return self._netloc['sdss']
        elif mirror or self.mirror:
            return self._netloc["mirror"]
        elif svn:
            return '{0}{1}'.format(self._netloc['svn'], '/public' if self.public else '')
        else:
            return self._netloc['sdss']

    def set_netloc(self, netloc=None, sdss=None, dtn=None, svn=None, mirror=None):
        ''' Set a url domain location

        Sets an SDSS url domain location.  Options are the SDSS SAS domain, the rsync download
        server, the svn server, or the mirror data domain.  The mirror data domain is set
        either by the ``mirror`` input keyword argument or by the ``path.mirror`` attribute.

        Parameters
        ----------
            netloc : str
                An exact net location to use directly
            sdss : bool
                If True, sets the SDSS data domain: data.sdss.org
            dtn : bool
                If True, sets the SDSS rsync server domain: dtn01.sdss.org
            svn: bool
                If True, sets the SDSS svn domain: svn.sdss.org
            mirror: bool
                If True, sets the SDSS mirror domain: data.mirror.sdss.org.

        '''
        self.netloc = self.get_netloc(netloc=netloc, sdss=sdss, dtn=dtn, svn=svn, mirror=mirror)

    def set_remote_base(self, scheme='https'):
        self.remote_base = self.get_remote_base(scheme=scheme or 'https')

    def get_remote_base(self, scheme="https", svn=None):
        ''' Get the remote base path

        Parameters
        ----------
        scheme : str
            The url scheme. Either "https" or "rsync".
        svn : bool
            If True, uses the svn url domain svn.sdss.org as the netloc
        '''
        netloc = self.netloc
        if svn:
            netloc = self.get_netloc(svn=True)
        return "{scheme}://{netloc}".format(scheme=scheme, netloc=netloc)

    def set_base_dir(self, base_dir=None):
        ''' Sets the base directory

        Sets the ``base_dir`` attribute.  Defaults to $SAS_BASE_DIR.  Can be
        overridden by passing in ``base_dir`` keyword argument.  The ``base_dir`` sets
        the beginning part of all local paths.

        Parameters
        ----------
            base_dir : str
                A directory path to use as the base

        '''
        if base_dir:
            self.base_dir = join(base_dir, '')
        else:
            try:
                self.base_dir = join(os.environ['SAS_BASE_DIR'], '')
            except Exception:
                pass

    @staticmethod
    def yield_product_root():
        ''' yields a product root environment name '''

        for root in tree._product_roots:
            yield root

    def find_location(self, filetype, **kwargs):
        ''' Finds a relative location of a product path

        Attempts to find a relative path location for a software product path.
        Loops over all product_roots defined in the tree and tests if a relative location
        can be extracted, i.e. if the path starts with a given root path.  The root environment
        paths searched are the following in order of precendence:
        PRODUCT_ROOT, SDSS_SVN_ROOT, SDSS_INSTALL_PRODUCT_ROOT, SDSS_PRODUCT_ROOT,
        SDSS4_PRODUCT_ROOT. If no root is found uses one directory up from SAS_BASE_DIR.

        Parameters
        ----------
        filetype : str
            File type parameter.
        kwargs : dict
            Path definition keyword arguments

        Returns
        -------
            The relative path location (to the base_dir)

        '''

        # loop over all potential git/svn product roots
        loc = None
        for root in tree._product_roots:
            loc = self._extract_location(filetype, base_dir=os.getenv(root), **kwargs)
            if loc:
                self.product_root = os.getenv(root)
                break
        return loc

    def _extract_location(self, filetype, base_dir=None, **kwargs):
        ''' Extracts the relative path location of the file

        Parameters
        ----------
        filetype : str
            File type parameter.
        base_dir : str
            A root directory to use as the base.  Defaults to SAS_BASE_DIR.

        Returns
        -------
            The relative path location (to the base_dir)

        '''
        full = kwargs.get('full', None)
        if not full:
            full = self.full(filetype, **kwargs)

        self.set_base_dir(base_dir=base_dir)
        location = full[len(self.base_dir):] if full and full.startswith(self.base_dir) else None
        return location

    def location(self, filetype, base_dir=None, **kwargs):
        """Return the location of the relative sas path of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.
        base_dir : str
            A root directory to use as the base.  Defaults to SAS_BASE_DIR.

        Returns
        -------
            The relative path location (to the base_dir)
        """

        # extract the location using SAS_BASE_DIR as the base
        location = self._extract_location(filetype, base_dir=base_dir, **kwargs)

        # attempt to find a product location
        if not location:
            location = self.find_location(filetype, **kwargs)

        if location and '//' in location:
            location = location.replace('//', '/')

        return location

    def url(self, filetype, base_dir=None, sasdir='sas', **kwargs):
        """Return the url of a given type of file.

        Parameters
        ----------
        filetype : str
            File type parameter.
        base_dir : str
            A root directory to use as the base.  Defaults to SAS_BASE_DIR.

        Returns
        -------
        full : str
            The sas url to the file.
        """

        # determine the remote domain location
        remote_base = self.remote_base
        full = self.full(filetype, skip_tag_check=True, **kwargs)
        # if not on the SAS, assume it is an SVN product path
        if not full.startswith(os.getenv("SAS_BASE_DIR")):
            remote_base = self.get_remote_base(svn=True)
            sasdir = ''

        # get the location and set the url
        location = self.location(filetype, skip_tag_check=True, base_dir=base_dir, **kwargs)
        if not location:
            raise AccessError('Cannot construct url.  A path.location could not extracted. ')

        # create the url path
        url = join(remote_base, sasdir, location) if remote_base and location else None
        if not is_posix:
            url = url.replace(sep, '/')

        # handle edge case when a full path is passed in as path.url('', full=full)
        # sanity check on svn tags
        if 'svn.sdss.org' in url:
            tag_match = re.search(r'tags/(v?[0-9._]+)', url)
            if not tag_match:
                url = re.sub(r'(/v?[0-9._]+/)', r'/tags\1', url, count=1)
        return url


def _expandvars(template):
    ''' Recursively run os.path.expandvars

    Recursively calls os.path.expandvars

    Parameters:
        template (str):
            sdss_access path template

    Return:
        A path template with expanded environment variables
    '''
    template = os.path.expandvars(template)
    if template.startswith('$'):
        # if the envvar isn't in os.environ, then exit
        envvar = template.split('/', 1)[0]
        if envvar[1:] not in os.environ:
            return template

        # recurse down
        return _expandvars(template)
    return template


class Path(BasePath):
    """Class for construction of paths in general.  Sets a particular template file.

    Parameters
    ----------
    release : str
        The release name, e.g. 'DR15', 'MPL-9'.
    public : bool
        If True, uses public urls.  Only needed for public data releases. Automatically set to True when release contains "DR".
    mirror : bool
        If True, uses the mirror data domain url.  Default is False.
    verbose: bool
        If True, turns on verbosity.  Default is False.
    force_modules: bool
        If True, forces software products to use any existing Module environment paths

    Attributes
    ----------
    templates : dict
        The set of templates read from the configuration file.

    """

    def __init__(self, release=None, public=False, mirror=False, verbose=False, force_modules=None):
        super(Path, self).__init__(release=release, public=public, mirror=mirror, verbose=verbose,
                                   force_modules=force_modules)

    def plateid6(self, filetype, **kwargs):
        """Print plate ID, accounting for 5-6 digit plate IDs.

        Parameters
        ----------
        filetype : str
            File type parameter.
        plateid : int or str
            Plate ID number.  Will be converted to int internally.

        Returns
        -------
        plateid6 : str
            Plate ID formatted to a string of 6 characters.
        """
        plateid = int(kwargs['plateid'])
        if plateid < 10000:
            return "{:0>6d}".format(plateid)
        else:
            return "{:d}".format(plateid)

    def platedir(self, filetype, **kwargs):
        """Returns plate subdirectory in :envvar:`PLATELIST_DIR` of the form: ``NNNNXX/NNNNNN``.

        Parameters
        ----------
        filetype : str
            File type parameter.
        plateid : int or str
            Plate ID number.  Will be converted to int internally.

        Returns
        -------
        platedir : str
            Plate directory in the format ``NNNNXX/NNNNNN``.
        """
        plateid = int(kwargs['plateid'])
        plateid100 = plateid // 100
        subdir = "{:0>4d}".format(plateid100) + "XX"
        return os.path.join(subdir, "{:0>6d}".format(plateid))

    def plategrp(self, filetype, **kwargs):
        ''' Returns plate group subdirectory

        Parameters
        ----------
        filetype : str
            File type parameter.
        plate : int or str
            Plate ID number.  Will be converted to int internally.

        Returns
        -------
        plategrp : str
            Plate group directory in the format ``NNNNXX``.

        '''

        plate = kwargs.get('plate', kwargs.get('plateid', None))
        if not plate:
            return 'XX'
        return '{:0>4d}XX'.format(int(plate) // 100)

    def spectrodir(self, filetype, **kwargs):
        """Returns :envvar:`SPECTRO_REDUX` or :envvar:`BOSS_SPECTRO_REDUX`
        depending on the value of `run2d`.

        Parameters
        ----------
        filetype : str
            File type parameter.
        run2d : int or str
            2D Reduction ID.

        Returns
        -------
        spectrodir : str
            Value of the appropriate environment variable.
        """
        if str(kwargs['run2d']) in ('26', '103', '104'):
            return os.environ['SPECTRO_REDUX']
        else:
            return os.environ['BOSS_SPECTRO_REDUX']

    def definitiondir(self, filetype, **kwargs):
        """Returns definition subdirectory in :envvar:`PLATELIST_DIR` of the form: ``NNNNXX``.

        Parameters
        ----------
        filetype : str
            File type parameter.
        designid : int or str
            Design ID number.  Will be converted to int internally.

        Returns
        -------
        definitiondir : str
            Definition directory in the format ``NNNNXX``.
        """

        designid = int(kwargs['designid'])
        designid100 = designid // 100
        subdir = "{:0>4d}".format(designid100) + "XX"
        return subdir


class AccessError(Exception):
    pass
